# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys
import numpy as np
import logging
from idtrackerai.list_of_blobs import ListOfBlobs
from idtrackerai.list_of_fragments import ListOfFragments
from idtrackerai.blob import Blob
if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.generate_light_groundtruth_blob_list")

class GroundTruthBlob(object):
    """Lighter blob objects.
    Attributes:
        identity (preferring the one assigned by the user, if it is not None)
        centroid
        pixels (pixels is stored to check the groundtruth in crossings)
    """
    def __init__(self, attributes_to_get = ['identity', 'assigned_identity',
                                            'used_for_training', 'accumulation_step',
                                            'centroid', 'pixels',
                                            'frame_number',
                                            'is_an_individual', 'is_a_crossing',
                                            'blob_index', 'fragment_identifier']):
        self.attributes = attributes_to_get

    def get_attribute(self, blob):
        for attribute in self.attributes:
            if attribute == 'identity':
                setattr(self, attribute, getattr(blob, 'final_identity'))
            else:
                setattr(self, attribute, getattr(blob, attribute))

class IndividualGroundTruth(object):
    def __init__(self, video = [], individual_blobs_in_video = [], start = None, end = None, validated_identity = None):
        self.video = video
        self.individual_blobs_in_video = individual_blobs_in_video
        self.start = start
        self.end = end
        self.validated_identity = validated_identity

    def save(self):
        gt_name = '_individual_' + str(self.validated_identity) + '_groundtruth.npy'
        path_to_save_groundtruth = os.path.join(os.path.split(self.video.video_path)[0], gt_name)
        logger.info("saving ground truth at %s" %path_to_save_groundtruth)
        np.save(path_to_save_groundtruth, self)
        logger.info("done")

def generate_individual_groundtruth(video, blobs_in_video = None,
                                    start = None, end = None,
                                    validated_identity = None, save_gt = True):
    """Generates a list of light blobs_in_video, given a video object corresponding to a
    tracked video
    """
    assert video.has_been_assigned == True
    individual_blobs_in_video_groundtruth = []

    for blobs_in_frame in blobs_in_video:
        identities_in_frame = set([blob.final_identity for blob in blobs_in_frame])
        for blob in blobs_in_frame:
            if blob.final_identity == validated_identity:
                gt_blob = GroundTruthBlob()
                gt_blob.get_attribute(blob)
                individual_blobs_in_video_groundtruth.append(gt_blob)

    groundtruth = IndividualGroundTruth(video = video,
                            individual_blobs_in_video = individual_blobs_in_video_groundtruth,
                            start = start,
                            end = end,
                            validated_identity = validated_identity)
    if save_gt:
        groundtruth.save()
    return groundtruth



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys
import itertools
import numpy as np
from tqdm import tqdm

from idtrackerai.blob import Blob
from idtrackerai.preprocessing.model_area import ModelArea
from idtrackerai.utils.py_utils import interpolate_nans

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.list_of_blobs")


class ListOfBlobs(object):
    """ Collects all the instances of the class :class:`~blob.Blob` generated
    from the blobs extracted from the video during segmentation
    (see :mod:`~segmentation`)

    Attributes
    ----------

    blobs_in_video : list
        List of instances of :class:`~blob.Blob` segmented from the video and
        organised framewise
    number_of_frames : int
        number of frames in the video
    blobs_are_connected :bool
        True if the blobs have already being organised in fragments (see
        :class:`~fragment.Fragment`). False otherwise
    """

    def __init__(self, blobs_in_video=None, number_of_frames=None):
        self.blobs_in_video = blobs_in_video
        self.number_of_frames = len(self.blobs_in_video)
        self.blobs_are_connected = False

    def disconnect(self):
        """Reinitialise the previous and next list of blobs
        (see :attr:`~blob.Blob.next` and :attr:`~blob.Blob.previous`)
        """
        for blobs_in_frame in self.blobs_in_video:
            for blob in blobs_in_frame:
                blob.next, blob.previous = [], []

    def connect(self):
        """Connects blobs in subsequent frames by computing their overlapping
        """
        logger.info("Connecting list of blob objects")
        for frame_i in tqdm(range(1, self.number_of_frames), desc='connecting blobs'):
            for (blob_0, blob_1) in itertools.product(self.blobs_in_video[frame_i-1], self.blobs_in_video[frame_i]):
                if blob_0.overlaps_with(blob_1):
                    blob_0.now_points_to(blob_1)

    def reconnect(self):
        """Connects blobs in subsequent frames by computing their overlapping
        and sets blobs_are_connected to True
        """
        logger.info("re-Connecting list of blob objects")
        for frame_i in tqdm(range(1, self.number_of_frames), desc='re-connecting blobs'):
            for (blob_0, blob_1) in itertools.product(self.blobs_in_video[frame_i-1], self.blobs_in_video[frame_i]):
                if blob_0.fragment_identifier == blob_1.fragment_identifier:
                    blob_0.now_points_to(blob_1)
        self.blobs_are_connected = True

    def save(self, video, path_to_save=None, number_of_chunks=1):
        """save instance"""
        self.disconnect()
        logger.info("saving blobs list at %s" % path_to_save)
        np.save(path_to_save, self)
        if video.has_been_segmented and not video.has_been_preprocessed:
            self.connect()
        self.blobs_are_connected = True

    @classmethod
    def load(cls, video, path_to_load_blob_list_file):
        """Short summary.

        Parameters
        ----------
        video : <Video object>
            See :class:`~video.Video`
        path_to_load_blob_list_file : str
            path to load a list of blobs

        Returns
        -------
        <ListOfBlobs object>
            an instance of ListOfBlobs

        """
        logger.info("loading blobs list from %s" % path_to_load_blob_list_file)
        list_of_blobs = np.load(path_to_load_blob_list_file, allow_pickle=True).item()
        list_of_blobs.blobs_are_connected = False
        return list_of_blobs

    def compute_fragment_identifier_and_blob_index(self, number_of_animals):
        """Associates a unique fragment identifier to the fragments computed by
        overlapping see method
        :meth:`compute_overlapping_between_subsequent_frames` and sets the blob
        index as the hierarchy of the blob in the first frame of the fragment to
        which blob belongs to.

        Parameters
        ----------
        number_of_animals : int
            number of animals to be tracked
        """
        counter = 0
        possible_blob_indices = range(number_of_animals)

        for blobs_in_frame in tqdm(self.blobs_in_video, desc='assigning fragment identifier'):
            used_blob_indices = [blob.blob_index for blob in blobs_in_frame if blob.blob_index is not None]
            missing_blob_indices = list(set(possible_blob_indices).difference(set(used_blob_indices)))
            for blob in blobs_in_frame:
                if blob.fragment_identifier is None and blob.is_an_individual:
                    blob._fragment_identifier = counter
                    blob_index = missing_blob_indices.pop(0)
                    blob._blob_index = blob_index
                    blob.non_shared_information_with_previous = 1.
                    if len(blob.next) == 1 and len(blob.next[0].previous) == 1 and blob.next[0].is_an_individual:
                        blob.next[0]._fragment_identifier = counter
                        blob.next[0]._blob_index = blob_index
                        # blob.next[0].compute_overlapping_with_previous_blob()
                        if blob.next[0].is_an_individual_in_a_fragment:
                            blob = blob.next[0]

                            while len(blob.next) == 1 and blob.next[0].is_an_individual_in_a_fragment:
                                blob = blob.next[0]
                                blob._fragment_identifier = counter
                                blob._blob_index = blob_index
                                # compute_overlapping_with_previous_blob
                                # blob.compute_overlapping_with_previous_blob()

                            if len(blob.next) == 1 and len(blob.next[0].previous) == 1 and blob.next[0].is_an_individual:
                                blob.next[0]._fragment_identifier = counter
                                blob.next[0]._blob_index = blob_index
                                # blob.next[0].compute_overlapping_with_previous_blob()
                    counter += 1

        self.number_of_individual_fragments = counter
        logger.info("number_of_individual_fragments, %i" %counter)

    def compute_crossing_fragment_identifier(self):
        """Assign a unique identifier to fragments associated to a crossing
        """
        def propagate_crossing_identifier(blob, fragment_identifier):
            """Propagates the identifier throughout the entire fragment to which
            blob belongs to. It crawls by overlapping in both the past and
            future "overlapping histories" of blob

            Parameters
            ----------
            blob : <Blob object>
                an instance of :class:`~blob.Blob`
            fragment_identifier : int
                unique fragment identifier associated
            """
            assert blob.fragment_identifier is None
            blob._fragment_identifier = fragment_identifier
            cur_blob = blob

            while len(cur_blob.next) == 1 and len(cur_blob.next[0].previous) == 1 and cur_blob.next[0].is_a_crossing:
                cur_blob = cur_blob.next[0]
                cur_blob._fragment_identifier = fragment_identifier

            cur_blob = blob

            while len(cur_blob.previous) == 1 and len(cur_blob.previous[0].next) == 1 and cur_blob.previous[0].is_a_crossing:
                cur_blob = cur_blob.previous[0]
                cur_blob._fragment_identifier = fragment_identifier

        fragment_identifier = self.number_of_individual_fragments

        for blobs_in_frame in self.blobs_in_video:
            for blob in blobs_in_frame:
                if blob.is_a_crossing and blob.fragment_identifier is None:
                    propagate_crossing_identifier(blob, fragment_identifier)
                    fragment_identifier += 1
        logger.info("number_of_crossing_fragments: %i" % (fragment_identifier - self.number_of_individual_fragments))
        logger.info("total number of fragments: %i" % fragment_identifier)

    def compute_overlapping_between_subsequent_frames(self):
        """Computes overlapping between self and the blobs generated during
        segmentation in the next frames
        """
        def set_frame_number_to_blobs_in_frame(blobs_in_frame, frame_number):
            """Sets the attribute frame number to the instances of
            :class:`~blob.Blob` collected in `blobs_in_frame`

            Parameters
            ----------
            blobs_in_frame : list
                list of instances of :class:`~blob.Blob`
            frame_number : int
                number of the frame from which `blobs_in_frame` has been
                generated
            """
            for blob in blobs_in_frame:
                blob.frame_number = frame_number
        self.disconnect()

        for frame_i in tqdm(range(1, self.number_of_frames), desc='Connecting blobs '):
            set_frame_number_to_blobs_in_frame(self.blobs_in_video[frame_i-1], frame_i-1)

            for (blob_0, blob_1) in itertools.product(self.blobs_in_video[frame_i-1], self.blobs_in_video[frame_i]):
                if blob_0.overlaps_with(blob_1):
                    blob_0.now_points_to(blob_1)

        set_frame_number_to_blobs_in_frame(self.blobs_in_video[frame_i], frame_i)

    def compute_model_area_and_body_length(self, number_of_animals):
        """computes the median and standard deviation of all the blobs of the video
        and the median_body_length estimated from the diagonal of the bounding box.
        These values are later used to discard blobs that are not fish and potentially
        belong to a crossing.
        """
        #areas are collected throughout the entire video in the cores of the global fragments
        areas_and_body_length = np.asarray([(blob.area,blob.estimated_body_length)
                                            for blobs_in_frame in self.blobs_in_video
                                            for blob in blobs_in_frame
                                            if len(blobs_in_frame) == number_of_animals])
        if areas_and_body_length.shape[0] == 0:
            raise ValueError('There is not part in the video where all the animals are visible. Try a different segmentation or check the number of animals in the video.')
        median_area = np.median(areas_and_body_length[:,0])
        mean_area = np.mean(areas_and_body_length[:,0])
        std_area = np.std(areas_and_body_length[:,0])
        median_body_length = np.median(areas_and_body_length[:,1])
        return ModelArea(mean_area, median_area, std_area), median_body_length

    def apply_model_area_to_video(self, video, model_area, identification_image_size, number_of_animals):
        """Applies `model_area` to every blob extracted from video

        Parameters
        ----------
        video : <Video object>
            See :class:`~video.Video`
        model_area : <ModelArea object>
            See :class:`~model_area.ModelArea`
        identification_image_size : int
            size of the identification image (see
            :meth:`~blob.Blob.get_image_for_identification` and
            :attr:`~blob.Blob.image_for_identification`)
        number_of_animals : int
            number of animals to be tracked
        """
        def apply_model_area_to_blobs_in_frame(video, number_of_animals,
                                               blobs_in_frame, model_area,
                                               identification_image_size):
            """Applies `model_area` to the collection of Blob instances in
            `blobs_in_frame`

            Parameters
            ----------
            video : <Video object>
                See :class:`~video.Video`
            number_of_animals : int
                number of animals to be tracked
            blobs_in_frame : list
                list of instances of :class:`~blob.Blob`
            model_area : <ModelArea object>
                See :class:`~model_area.ModelArea`
            identification_image_size : int
                size of the identification image (see
                :meth:`~blob.Blob.get_image_for_identification` and
                :attr:`~blob.Blob.image_for_identification`)
            """
            number_of_blobs = len(blobs_in_frame)
            for blob in blobs_in_frame:
                blob.apply_model_area(video, number_of_animals, model_area,
                                      identification_image_size,
                                      number_of_blobs)
        for blobs_in_frame in tqdm(self.blobs_in_video, desc='Applying model area'):
            apply_model_area_to_blobs_in_frame(video, number_of_animals,
                                               blobs_in_frame, model_area,
                                               identification_image_size)

    # def get_data_plot(self):
    #     """Gets the areas of all the blobs segmented in the video
    #
    #     Returns
    #     -------
    #     list
    #         Areas of all the blobs segmented in the video
    #
    #     """
    #     return [blob.area for blobs_in_frame in self.blobs_in_video for blob in blobs_in_frame]


    def check_maximal_number_of_blob(self, number_of_animals, return_maximum_number_of_blobs = False):
        """Checks that the amount of blobs per frame is not greater than the
        number of animals to track

        Parameters
        ----------
        number_of_animals : int
            number of animals to be tracked

        Returns
        -------
        list
            List of indices of frames in which more blobs than animals to track
            have been segmented

        """
        maximum_number_of_blobs = 0
        frames_with_more_blobs_than_animals = []
        for frame_number, blobs_in_frame in enumerate(self.blobs_in_video):

            if len(blobs_in_frame) > number_of_animals:
                frames_with_more_blobs_than_animals.append(frame_number)
            maximum_number_of_blobs = len(blobs_in_frame) if len(blobs_in_frame) > maximum_number_of_blobs else maximum_number_of_blobs

        if len(frames_with_more_blobs_than_animals) > 0:
            logger.error('There are frames with more blobs than animals, this can be detrimental for the proper functioning of the system.')
            logger.error("Frames with more blobs than animals: %s" %str(frames_with_more_blobs_than_animals))
        if return_maximum_number_of_blobs:
            return frames_with_more_blobs_than_animals, maximum_number_of_blobs
        else:
            return frames_with_more_blobs_than_animals


    def update_from_list_of_fragments(self, fragments, fragment_identifier_to_index):
        """Updates the blobs objects generated from the video with the attributes
        computed for each fragment

        Parameters
        ----------
        fragments : list
            List of all the fragments
        fragment_identifier_to_index : int
            index to retrieve the fragment corresponding to a certain fragment
            identifier (see
            :meth:`~blob.Blob.compute_fragment_identifier_and_blob_index`)

        """
        attributes = ['identity',
                        'P2_vector',
                        'identity_corrected_solving_jumps',
                        'user_generated_identity', 'used_for_training',
                        'accumulation_step']

        for blobs_in_frame in tqdm(self.blobs_in_video, desc = 'updating list of blobs from list of fragments'):
            for blob in blobs_in_frame:
                fragment = fragments[fragment_identifier_to_index[blob.fragment_identifier]]
                [setattr(blob, '_' + attribute, getattr(fragment, attribute)) for attribute in attributes if hasattr(fragment, attribute)]

    def next_frame_to_validate(self, current_frame, direction):
        """Returns the next frame to be validated given the current frame an
        the direction to iterate

        Parameters
        ----------
        current_frame : int
            Frame from which to start checking for frames to validate
        direction : string
            Direction towards where to start checking. 'future' will check for
            upcoming frames, and 'past' for previous frames.

        Returns
        -------
        frame_number : int

        """
        logger.debug('next_frame_to_validate: {0}'.format(current_frame))

        if not (current_frame > 0 and current_frame < len(self.blobs_in_video)):
            raise Exception("The frame number must be between 0 and the number of frames in the video")
        if direction == 'future':
            blobs_in_frame_to_check = self.blobs_in_video[current_frame+1:]
        elif direction == 'past':
            blobs_in_frame_to_check = self.blobs_in_video[0:current_frame][::-1]
        for blobs_in_frame in blobs_in_frame_to_check:
            for blob in blobs_in_frame:
                if check_tracking(blobs_in_frame):
                    return blob.frame_number

    def interpolate_from_user_generated_centroids(self, video, identity,
                                                  start_frame, end_frame):
        """
        Interpolates linearly the centroids of the blobs of identity identity between
        start_frame and end_frame. The interpolation is done using the
        user_generated_centroids. The centroid of the blobs without
        user_generated_centroids are assumed to be nan and are interpolated
        accordingly.

        Parameters
        ----------
        identity : int
            Identity of the blobs to be interpolated
        start_frame : int
            Frame from which to start interpolation
        end_frame : int
            Frame where to end the interpolation
        """
        end_frame = end_frame + 1
        if start_frame >= end_frame:
            raise Exception('The initial frame has to be higher than the last frame.')

        # Collect centroids of blobs with identity identity that were modified by the user
        centroids_to_interpolate = []
        blobs_of_id = []
        for blobs_in_frame in self.blobs_in_video[start_frame:end_frame]:
            possible_blobs = [blob for blob in blobs_in_frame if identity in blob.final_identities]

            if len(possible_blobs) == 1:
                blobs_of_id.append(possible_blobs[0])
                identity_index = possible_blobs[0].final_identities.index(identity)
                fixed_centroid = (None, None)
                if possible_blobs[0].user_generated_centroids is not None:
                    fixed_centroid = possible_blobs[0].user_generated_centroids[identity_index]
                if (fixed_centroid[0] is not None and fixed_centroid[0] > 0):
                    centroids_to_interpolate.append(fixed_centroid)
                elif possible_blobs[0].user_generated_identities is not None and possible_blobs[0].user_generated_identities[identity_index] is not None:
                    centroids_to_interpolate.append(possible_blobs[0].final_centroids[identity_index])
                else:
                    centroids_to_interpolate.append((np.nan, np.nan))
            elif not possible_blobs:
                blobs_of_id.append(None)
                centroids_to_interpolate.append((np.nan, np.nan))
            else:
                raise Exception('Make sure that the identnties of the user \
                                generated centroids (marked with u-) are unique \
                                in the interpolation interval.')

        try:
            if not (len(centroids_to_interpolate) == len(blobs_of_id) == end_frame - start_frame):
                raise Exception("The number of user generated centroids before interpolation does \
                                not match the number of frames interpolated.")
            if np.isnan(centroids_to_interpolate[0][0]) or np.isnan(centroids_to_interpolate[-1][0]):
                raise Exception("The first and last frame of the interpolation interval must contain a \
                                user generated centroid marked with the 'u-' prefix")
            centroids_to_interpolate = np.asarray(centroids_to_interpolate)
            # interpolate linearlry the centroids not generated by the user
            interpolate_nans(centroids_to_interpolate)
            # assign the new centroids to the blobs with identity identity.
            frames = range(start_frame, end_frame)
            for i, (blob, frame) in enumerate(zip(blobs_of_id, frames)):
                if blob is not None:
                    identity_index = blob.final_identities.index(identity)
                    if blob._user_generated_centroids is None:
                        blob._user_generated_centroids = [(None, None)]*len(blob.final_centroids)
                    blob.user_generated_centroids[identity_index] = tuple(centroids_to_interpolate[i, :])
                else:
                    blob_index = np.argmin([candidate_blob.distance_from_countour_to(tuple(centroids_to_interpolate[i,:]))
                                            for candidate_blob in self.blobs_in_video[frame]
                                            if candidate_blob.is_a_crossing])
                    nearest_blob = self.blobs_in_video[frame][blob_index]
                    nearest_blob.add_centroid(video, tuple(centroids_to_interpolate[i,:]), identity,
                                              apply_resolution_reduction=False)
        except:
            raise Exception('Something went wrong during the interpolation')

    def reset_user_generated_identities_and_centroids(self, video, start_frame,
                                                      end_frame, identity=None):
        """
        Resets the identities and centroids generetad by the user to the ones assigned by the software

        Parameters
        ----------
        start_frame : int
            Frame from which to start reseting identities and centroids
        end_frame : int
            Frame where to end reseting identities and centroids
        identity : int, optional
            Identity of the blobs to be reseted (default None). If None,
            all the blobs are reseted
        """
        if start_frame > end_frame:
            raise Exception("Initial frame number must be smaller than the final frame number")
        if not (identity is None or identity > 0):
            # missing identity <= self.number_of_animals but the attribute does not exist
            raise Exception("Identity must be None or a positive integer")

        for blobs_in_frame in self.blobs_in_video[start_frame:end_frame]:
            if identity is None:
                # Reset all user generated identities and centroids
                for blob in blobs_in_frame:
                    blob._user_generated_identities = None
                    blob._user_generated_centroids = None
            else:
                possible_blobs = [blob for blob in blobs_in_frame
                                  if identity in blob.final_identities]
                for blob in possible_blobs:
                    indices = [i for i, final_id in enumerate(blob.final_identities)
                               if final_id == identity]
                    for index in indices:
                        if blob._user_generated_centroids is not None:
                            blob._user_generated_centroids[index] = (None, None)
                        if blob._user_generated_identities is not None:
                            blob._user_generated_identities[index] = None

        video._is_centroid_updated = any([any([cent[0] is not None
                                               for cent in blob.user_generated_centroids])
                                          for blobs_in_frame in self.blobs_in_video
                                          for blob in blobs_in_frame
                                          if blob.user_generated_centroids is not None])


def check_tracking(blobs_in_frame):
    """Returns True if the list of blobs blobs_in_frame needs to be checked.

    Parameters
    ----------
    blobs_in_frame : list
        List of Blob objects

    Returns
    -------
    check_tracking_flag : boolean


    """
    there_are_crossings = any([b.is_a_crossing for b in blobs_in_frame]) #check whether there is a crossing in the frame
    missing_identity = any([len(b.final_identities) == 1 and b.final_identities[0] == 0
                            for b in blobs_in_frame]) # Check whether there is some missing identities (0 or None)
    return there_are_crossings or missing_identity



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os

import numpy as np
import cv2
import h5py

from idtrackerai.blob import Blob
from idtrackerai.utils.segmentation_utils import blob_extractor

''' erosion '''


def compute_erosion_disk(video, blobs_in_video):
    return np.ceil(np.nanmedian([compute_min_frame_distance_transform(video, blobs_in_frame)
                                 for blobs_in_frame in blobs_in_video
                                 if len(blobs_in_frame) > 0])).astype(np.int)


def compute_min_frame_distance_transform(video, blobs_in_frame):
    max_distance_transform = [compute_max_distance_transform(video, blob)
                              for blob in blobs_in_frame
                              if blob.is_an_individual]
    return np.min(max_distance_transform) if len(max_distance_transform) > 0 else np.nan


def generate_temp_image(video, pixels, bounding_box_in_frame_coordinates):
    pxs = np.array(np.unravel_index(pixels,(video.height, video.width))).T
    pxs = np.array([pxs[:, 0] - bounding_box_in_frame_coordinates[0][1],
                    pxs[:, 1] - bounding_box_in_frame_coordinates[0][0]])
    temp_image = np.zeros((bounding_box_in_frame_coordinates[1][1] -
                            bounding_box_in_frame_coordinates[0][1],
                            bounding_box_in_frame_coordinates[1][0] -
                            bounding_box_in_frame_coordinates[0][0])).astype('uint8')
    temp_image[pxs[0,:], pxs[1,:]] = 255
    return temp_image


def compute_max_distance_transform(video, blob):
    temp_image = generate_temp_image(video, blob.pixels, blob.bounding_box_in_frame_coordinates)
    return np.max(cv2.distanceTransform(temp_image, cv2.DIST_L2, cv2.DIST_MASK_PRECISE))

def erode(image, kernel_size):
    kernel = np.ones(kernel_size, np.uint8)
    return cv2.erode(image,kernel,iterations = 1)

def get_eroded_blobs(video, blobs_in_frame, frame_number):
    episode = video.in_which_episode(frame_number)
    pixels_path = None
    if conf.SAVE_PIXELS == 'DISK':
        pixels_path = os.path.join(video._segmentation_data_folder,
                                   'episode_pixels_{}.hdf5'.format(str(episode)))

    # logger.debug('Getting eroded blobs')
    segmented_frame = np.zeros((video.height, video.width)).astype('uint8')

    for blob in blobs_in_frame:
        pixels = blob.eroded_pixels if (hasattr(blob, 'has_eroded_pixels') and blob.has_eroded_pixels) else blob.pixels
        pxs = np.array(np.unravel_index(pixels,(video.height, video.width))).T
        segmented_frame[pxs[:,0], pxs[:,1]] = 255

    segmented_eroded_frame = erode(segmented_frame, video.erosion_kernel_size)
    boundingBoxes, _, centroids, _, pixels_all, contours, _ = blob_extractor(segmented_eroded_frame, segmented_eroded_frame, 0, np.inf)
    # logger.debug('Finished getting eroded blobse')
    eroded_blobs_in_frame = []
    for i, (centroid, contour, pixels, bounding_box) in enumerate(zip(centroids, contours, pixels_all, boundingBoxes)):
        if conf.SAVE_PIXELS == 'DISK':
            with h5py.File(pixels_path, 'a') as f:
                f.create_dataset(str(frame_number) + '-' + str(i) + '-eroded',
                                  data=pixels)
            pixels=None

        eroded_blobs_in_frame.append(Blob(centroid,
                                          contour,
                                          None,
                                          bounding_box,
                                          bounding_box_image=None,
                                          number_of_animals=video.number_of_animals,
                                          frame_number=frame_number,
                                          pixels=pixels,
                                          pixels_path=pixels_path,
                                          in_frame_index=i,
                                          video_height=video.height,
                                          video_width=video.width,
                                          video_path=video.video_path,
                                          has_eroded_pixels=True,
                                          resolution_reduction=video.resolution_reduction))

    return eroded_blobs_in_frame



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

import matplotlib.pyplot as plt
import numpy as np

from idtrackerai.network.crossings_detector_model.stop_training_criteria_crossings import Stop_Training
from idtrackerai.network.crossings_detector_model.store_accuracy_and_loss_crossings import Store_Accuracy_and_Loss
from idtrackerai.network.crossings_detector_model.epoch_runner_crossings import EpochRunner

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.train_crossing_detector")


class TrainDeepCrossing(object):
    def __init__(self, net, training_dataset, validation_dataset,
                 num_epochs=50, plot_flag=True, return_store_objects=False):
        """Build the dataset and trains the model
        The dataset is built according to
        Taken from tensorflow/contrib/learn/datasets/mnist.
        """
        self.training_dataset = training_dataset
        self.validation_dataset = validation_dataset
        self.num_epochs = num_epochs
        self.plot_flag = plot_flag
        self.net = net
        self.return_store_objects = return_store_objects
        self.train_model()

    def train_model(self, global_step=0,
                    check_for_loss_plateau=None,
                    print_flag=False,
                    store_accuracy_and_error=False):
        logger.info("\nTraining Deep Crossing Detector")
        store_training_accuracy_and_loss_data = Store_Accuracy_and_Loss(self.net, name='training')
        store_validation_accuracy_and_loss_data = Store_Accuracy_and_Loss(self.net, name='validation')

        if self.plot_flag:
            # Initialize pre-trainer plot
            plt.ion()
            fig, ax_arr = plt.subplots(3)
            title = 'Crossing-detector'
            fig.canvas.set_window_title(title)
            fig.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.5)

        self.net.compute_loss_weights(self.training_dataset.labels)
        trainer = EpochRunner(self.training_dataset,
                              starting_epoch=global_step,
                              print_flag=print_flag)
        validator = EpochRunner(self.validation_dataset,
                                starting_epoch=global_step,
                                print_flag=print_flag)
        # set criteria to stop the training
        stop_training = Stop_Training(check_for_loss_plateau=check_for_loss_plateau)
        logger.debug("entering the epochs loop...")
        while not stop_training(store_training_accuracy_and_loss_data,
                                store_validation_accuracy_and_loss_data,
                                trainer._epochs_completed):
            # --- Training
            trainer.run_epoch('Training',
                              store_training_accuracy_and_loss_data,
                              self.net.train)
            # --- Validation
            validator.run_epoch('Validation',
                                store_validation_accuracy_and_loss_data,
                                self.net.validate)
            # Update counter
            trainer._epochs_completed += 1
            validator._epochs_completed += 1

        if (np.isnan(store_training_accuracy_and_loss_data.loss[-1]) or np.isnan(store_validation_accuracy_and_loss_data.loss[-1])):
            logger.warn("The model diverged. Falling back to individual-crossing discrimination by average area model.")
            self.model_diverged = True
        else:
            self.model_diverged = False
            global_step += trainer.epochs_completed
            logger.debug('validation losses: %s' % str(store_validation_accuracy_and_loss_data.loss))
            # plot if asked
            if self.plot_flag:
                store_training_accuracy_and_loss_data.plot(ax_arr, color='r')
                store_validation_accuracy_and_loss_data.plot(ax_arr, color='b')
                fig.savefig(os.path.join(self.net.params.save_folder, title + '.pdf'))
            # store training and validation losses and accuracies
            if store_accuracy_and_error:
                store_training_accuracy_and_loss_data.save()
                store_validation_accuracy_and_loss_data.save()
            # Save network model
            self.net.save(global_step=global_step)
            if self.return_store_objects:
                self.store_training_accuracy_and_loss_data = store_training_accuracy_and_loss_data
                self.store_validation_accuracy_and_loss_data = store_validation_accuracy_and_loss_data



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import numpy as np
from confapp import conf

from idtrackerai.globalfragment import GlobalFragment
from idtrackerai.assigner import assign, compute_identification_statistics_for_non_accumulated_fragments
from idtrackerai.accumulation_manager import AccumulationManager

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.list_of_global_fragments")

class ListOfGlobalFragments(object):
    """ Collects all the instances of the class
    :class:`~glboalfragment.GlobalFragment`
    generated by considering the fragments
    (see :meth:`~list_of_blobs.compute_fragment_identifier_and_blob_index`)

    Attributes
    ----------

    global_fragments : list
        list of instances of the class :class:`~globalfragment.GlobalFragment`
    number_of_global_fragments :  int
        length of `global_fragments`
    first_global_fragment_for_accumulation :  <GlobalFragment object>
        first global fragment used to start the protocols.
        See :meth:`set_first_global_fragment_for_accumulation`
    maximum_number_of_images : int
        maximum number of images contained in the global fragments listed in
        `global_fragments`
    non_accumulable_global_fragments : list
        list of global fragments that cannot be used during the accumulation
        protocols.
        See :attr:`~globalfragment.GlobalFragment.candidate_for_accumulation`

    Notes
    -----
    A minimum number of images per individual fragments is required for a
    global fragment to be acceptable for the accumulation. This number is set
    in
    :const:`conf.MINIMUM_NUMBER_OF_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION`

    """
    def __init__(self, global_fragments):
        self.global_fragments = global_fragments
        self.number_of_global_fragments = len(self.global_fragments)

    def reset(self, roll_back_to = None):
        """Resets all the global fragment by calling recursively the method
        :meth:`~globalfragment.GlobalFragment.reset`
        """
        for global_fragment in self.global_fragments:
            global_fragment.reset(roll_back_to)

    def order_by_distance_travelled(self):
        """Sorts the global fragments by the minimum distance travelled of their
        individual fragments
        """
        self.global_fragments = sorted(self.global_fragments, key = lambda x: x.minimum_distance_travelled, reverse = True)

    @staticmethod
    def give_me_frequencies_first_fragment_accumulated(i, number_of_animals, fragment):
        """The frequencies (see :meth:`~fragments.Fragments.compute_identification_statistics`)
        are generated artificially for the first global fragments.

        Parameters
        ----------
        i : int
            identity associated to the `fragment`
        number_of_animals : int
            number of animals to track
        fragment : <Fragment object>
            an instance of the class :class:`~fragment.Fragment`

        Returns
        -------

        ndarray
            array of zeros with the `i`th component equal to
            :attr:`~fragment.Fragment.number_of_images`
        """
        frequencies = np.zeros(number_of_animals)
        frequencies[i] = fragment.number_of_images
        return frequencies

    @staticmethod
    def abort_knowledge_transfer_on_same_animals(video, net):
        identities = range(video.number_of_animals)
        net.reinitialize_softmax_and_fully_connected()
        logger.info("Identity transfer failed. We proceed by transferring only the convolutional filters.")
        return identities

    def set_first_global_fragment_for_accumulation(self, video, accumulation_trial=0, net=None):
        """Selects the first global fragment to be used for accumulation

        Parameters
        ----------
        video : <Video object>
            instance of the class :class:`~video.Video`
        accumulation_trial : int
            accumulation number (protocol 2 performs a single accumulation
            attempt, and if used, protocol 3 will perform 3 other attempts)

        Returns
        -------
        int
            frame number corresponding to the beginning of the selected global
            fragment

        """
        self.order_by_distance_travelled()

        try:
            self.first_global_fragment_for_accumulation = self.global_fragments[accumulation_trial]
        except:
            return None

        if not video.identity_transfer or net is None:
            identities = range(video.number_of_animals)
        else:
            logger.info("Transferring identities from {}".format(video.knowledge_transfer_model_folder))
            identities = self.get_transferred_identities(video, net)


        [(setattr(fragment, '_acceptable_for_training', True),
            setattr(fragment, '_temporary_id', identities[i]),
            setattr(fragment, '_frequencies', self.give_me_frequencies_first_fragment_accumulated(i, video.number_of_animals, fragment)),
            setattr(fragment, '_is_certain', True),
            setattr(fragment, '_certainty', 1.),
            setattr(fragment, '_P1_vector', fragment.compute_P1_from_frequencies(fragment.frequencies)))
            for i, fragment in enumerate(self.first_global_fragment_for_accumulation.individual_fragments)]

        return self.first_global_fragment_for_accumulation.index_beginning_of_fragment

    def order_by_distance_to_the_first_global_fragment_for_accumulation(self, video, accumulation_trial = None):
        """Sorts the global fragments wrt to their distance from the first
        global fragment chose for accumulation

        Parameters
        ----------
        video : <Video object>
            instance of the class :class:`~video.Video`
        accumulation_trial : int
            accumulation number (protocol 2 performs a single accumulation
            attempt, and if used, protocol 3 will perform 3 other attempts)

        """
        self.global_fragments = sorted(self.global_fragments,
                                        key = lambda x: np.abs(x.index_beginning_of_fragment - video.first_frame_first_global_fragment[accumulation_trial]),
                                        reverse = False)

    def get_transferred_identities(self, video, net):
        """Assigns an identity to the images of the first global fragment using
        a network passed by the user to perform identity transfer

        Parameters
        ----------
        video : <Video object>
            instance of the class :class:`~video.Video`.
        net : ConvNetwork object
            network used to assign the identities of the first global fragment.

        """
        images, _ = self.first_global_fragment_for_accumulation.get_images_and_labels(video.identification_images_file_path, scope='identity_transfer')
        images = np.asarray(images)
        assigner = assign(net, images, False)
        compute_identification_statistics_for_non_accumulated_fragments(
            self.first_global_fragment_for_accumulation.individual_fragments,
            assigner, net.params.number_of_animals)
        # Check certainties of the individual fragments in the global fragment
        # for individual_fragment_identifier in global_fragment.individual_fragments_identifiers:
        [setattr(fragment, '_acceptable_for_training', True) for fragment
         in self.first_global_fragment_for_accumulation.individual_fragments]

        for fragment in self.first_global_fragment_for_accumulation.individual_fragments:
            if AccumulationManager.is_not_certain(fragment, conf.CERTAINTY_THRESHOLD):
                logger.debug('Identity transfer failed because a fragment is not cergain enough')
                logger.debug('CERTAINTY_THRESHOLD %.2f, fragment certainty %.2f' %(conf.CERTAINTY_THRESHOLD, fragment.certainty))
                identities = self.abort_knowledge_transfer_on_same_animals(video, net)
                return identities

        P1_array, index_individual_fragments_sorted_by_P1_max_to_min = AccumulationManager.get_P1_array_and_argsort(
                                self.first_global_fragment_for_accumulation)

        # assign temporary identity to individual fragments by hierarchical P1
        for index_individual_fragment in index_individual_fragments_sorted_by_P1_max_to_min:
            fragment = self.first_global_fragment_for_accumulation.individual_fragments[index_individual_fragment]

            if AccumulationManager.p1_below_random(P1_array, index_individual_fragment, fragment):
                logger.debug('Identity transfer failed because P1 is below random')
                identities = self.abort_knowledge_transfer_on_same_animals(video, net)
                return identities
            else:
                temporary_id = np.argmax(P1_array[index_individual_fragment,:])
                if not fragment.check_consistency_with_coexistent_individual_fragments(temporary_id):
                    logger.debug('Identity transfer failed because the identities are not consistent')
                    identities = self.abort_knowledge_transfer_on_same_animals(video, net)
                    return identities
                else:
                    P1_array = AccumulationManager.set_fragment_temporary_id(
                                            fragment, temporary_id, P1_array,
                                            index_individual_fragment)

        # Check if the global fragment is unique after assigning the identities
        if not self.first_global_fragment_for_accumulation.is_unique:
            logger.debug('Identity transfer failed because the identities are not unique')
            identities = self.abort_knowledge_transfer_on_same_animals(video, net)
            logger.info("Identity transfer is not possible. Identities will be intialized")
        else:
            video._first_global_fragment_knowledge_transfer_identities = [fragment.temporary_id for fragment
                        in self.first_global_fragment_for_accumulation.individual_fragments]
            if video.number_of_animals == video.knowledge_transfer_info_dict['number_of_animals']:
                identities = video._first_global_fragment_knowledge_transfer_identities
            elif video.number_of_animals < video.knowledge_transfer_info_dict['number_of_animals']:
                identities = range(video.number_of_animals)
            logger.info("Identities transferred succesfully")

        # self.plot_P1s_identity_transfer(video)
        return identities

    def compute_maximum_number_of_images(self):
        """Computes the maximum number of images in the global fragments
        """
        self.maximum_number_of_images = max([global_fragment.get_total_number_of_images() for global_fragment in self.global_fragments])

    def filter_candidates_global_fragments_for_accumulation(self):
        """Filters the global fragments by taking into account the minium
        number of images per individual fragments specified in
        :attr:`~globalfragment.GlobalFragment.candidate_for_accumulation`
        """
        self.non_accumulable_global_fragments = [global_fragment for global_fragment in self.global_fragments
                    if not global_fragment.candidate_for_accumulation]
        self.global_fragments = [global_fragment for global_fragment in self.global_fragments
                    if global_fragment.candidate_for_accumulation]
        self.number_of_global_fragments = len(self.global_fragments)

    # def get_data_plot(self):
    #     """Gathers data to plot a global fragments statistics summary
    #
    #     Returns
    #     -------
    #     int
    #         number of images in the shortest individual fragment
    #     int
    #         number of images in the longest individual fragment
    #     int
    #         number of images per individual fragment in global fragment
    #     int
    #         median number of images in global fragments
    #     float
    #         minimum distance travelled
    #
    #     """
    #     number_of_images_in_shortest_individual_fragment = []
    #     number_of_images_in_longest_individual_fragment = []
    #     number_of_images_per_individual_fragment_in_global_fragment = []
    #     median_number_of_images = []
    #     minimum_distance_travelled = []
    #     for global_fragment in self.global_fragments:
    #         number_of_images_in_shortest_individual_fragment.append(min(global_fragment.number_of_images_per_individual_fragment))
    #         number_of_images_in_longest_individual_fragment.append(max(global_fragment.number_of_images_per_individual_fragment))
    #         number_of_images_per_individual_fragment_in_global_fragment.append(global_fragment.number_of_images_per_individual_fragment)
    #         median_number_of_images.append(np.median(global_fragment.number_of_images_per_individual_fragment))
    #         minimum_distance_travelled.append(min(global_fragment.distance_travelled_per_individual_fragment))
    #
    #     return number_of_images_in_shortest_individual_fragment,\
    #             number_of_images_in_longest_individual_fragment,\
    #             number_of_images_per_individual_fragment_in_global_fragment,\
    #             median_number_of_images,\
    #             minimum_distance_travelled

    def delete_fragments_from_global_fragments(self):
        """Deletes the individual fragments from each of the global fragments
        """
        for global_fragment in self.global_fragments:
            global_fragment.individual_fragments = None

    def relink_fragments_to_global_fragments(self, fragments):
        """Resets the individual fragments to their respective global fragments
        """
        for global_fragment in self.global_fragments:
            global_fragment.get_individual_fragments_of_global_fragment(fragments)

    def save(self, global_fragments_path, fragments):
        """Saves an instance of the class in the path `global_fragments_path`.
        Before saving the individual fragments associated to every global
        fragment are removes by using the method
        :meth:`~delete_fragments_from_global_fragments`
        and resets them after saving by calling
        :meth:`~relink_fragments_to_global_fragments`
        """
        logger.info("saving list of global fragments at %s" %global_fragments_path)
        self.delete_fragments_from_global_fragments()
        np.save(global_fragments_path,self)
        # After saving the list of globa fragments the individual fragments are deleted and we need to relink them again
        self.relink_fragments_to_global_fragments(fragments)

    @classmethod
    def load(self, path_to_load, fragments):
        """Loads an instance of the class saved with :meth:`save` and
        associates individual fragments to each global fragment by calling
        :meth:`~relink_fragments_to_global_fragments`
        """
        logger.info("loading list of global fragments from %s" %path_to_load)
        list_of_global_fragments = np.load(path_to_load, allow_pickle=True).item()
        list_of_global_fragments.relink_fragments_to_global_fragments(fragments)
        return list_of_global_fragments

def detect_beginnings(boolean_array):
    """ Detects the frame where the core of a global fragment starts.
    A core of a global fragment is the part of the global fragment where all the
    individuals are visible, i.e. the number of animals in the frame equals the
    number of animals in the video
    :boolean_array: array with True where the number of animals in the frame equals
    the number of animals in the video
    """
    if np.all(boolean_array):
        return [0]
    else:
        return [i for i in range(0,len(boolean_array)) if (boolean_array[i] and not boolean_array[i-1])]

def check_global_fragments(blobs_in_video, num_animals):
    """Returns an array with True if:
    * each blob has a unique blob intersecting in the past and future
    * number of blobs equals num_animals
    """
    def all_blobs_in_a_fragment(blobs_in_frame):
        """Returns all the blobs in `blobs_in_frame` that are associated to an
        individual
        """
        return all([blob.is_an_individual for blob in blobs_in_frame])

    def same_fragment_identifier(blobs_in_frame, blobs_in_frame_past):
        """Return True if the set of fragments identifiers in the current frame
        is the same as in the previous frame, otherwise returns false
        """
        condition_1 = set([blob.fragment_identifier for blob in blobs_in_frame]) == set([blob.fragment_identifier for blob in blobs_in_frame_past])
        condition_2 = all_blobs_in_a_fragment(blobs_in_frame_past) and len(blobs_in_frame_past) == num_animals
        return condition_1 or not condition_2

    return [all_blobs_in_a_fragment(blobs_in_frame)
            and len(blobs_in_frame) == num_animals
            and same_fragment_identifier(blobs_in_frame, blobs_in_video[i-1])
            for i, blobs_in_frame in enumerate(blobs_in_video)]

def create_list_of_global_fragments(blobs_in_video, fragments, num_animals):
    """Creates the list of instances of the class
    :class:`~globalfragment.GlobalFragment` used to create :class:`.ListOfGlobalFragments`

    Parameters
    ----------
    blobs_in_video : list
        list of the blob objects (see class :class:`~blob.Blob`) generated
        from the blobs segmented in the video
    fragments : list
        list of instances of the class :class:`~fragment.Fragment`
    num_animals : int
        number of animals to track

    Returns
    -------
    list
        list of instances of the class :class:`~globalfragment.GlobalFragment`

    """
    global_fragments_boolean_array = check_global_fragments(blobs_in_video, num_animals)
    indices_beginning_of_fragment = detect_beginnings(global_fragments_boolean_array)
    return [GlobalFragment(blobs_in_video, fragments, i, num_animals)
            for i in indices_beginning_of_fragment]



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

import collections
import numpy as np
import cv2
from tqdm import tqdm
from scipy.spatial.distance import cdist
from confapp import conf

from idtrackerai.preprocessing.erosion import compute_erosion_disk, \
                                            erode, \
                                            get_eroded_blobs
from idtrackerai.list_of_fragments import ListOfFragments
from idtrackerai.list_of_blobs import ListOfBlobs
from idtrackerai.postprocessing.compute_velocity_model import compute_model_velocity


if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.assign_them_all")

''' assign them all '''
def set_individual_with_identity_0_as_crossings(list_of_blobs_no_gaps):
    for blobs_in_frame in list_of_blobs_no_gaps.blobs_in_video:
        for blob in blobs_in_frame:
            if blob.is_an_individual and len(blob.assigned_identities) == 1 and blob.assigned_identities[0] == 0:
                blob._is_an_individual = False
                blob._is_a_crossing = True
                blob._identity = None
                blob._identity_corrected_solving_jumps = None


def find_the_gap_interval(blobs_in_video, possible_identities, gap_start, list_of_occluded_identities):
    # logger.debug('Finding gap interval')
    there_are_missing_identities = True
    frame_number = gap_start + 1
    if frame_number < len(blobs_in_video):

        while there_are_missing_identities and frame_number > 0 and frame_number < len(blobs_in_video):
            blobs_in_frame = blobs_in_video[frame_number]
            occluded_identities_in_frame = list_of_occluded_identities[frame_number]
            missing_identities = get_missing_identities_from_blobs_in_frame(possible_identities, blobs_in_frame, occluded_identities_in_frame)
            if len(missing_identities) == 0 or frame_number == len(blobs_in_video)-1:
                there_are_missing_identities = False
            else:
                frame_number += 1
            gap_end = frame_number
    else:
        gap_end = gap_start
    # logger.debug('Finished finding gap interval')
    return (gap_start, gap_end)


def get_blob_by_identity(blobs_in_frame, identity):
    for blob in blobs_in_frame:
        if identity in blob.final_identities:
            return [blob]
    return None


def get_candidate_blobs_by_overlapping(blob_to_test, eroded_blobs_in_frame):
    # logger.debug('Getting candidate blobs by overlapping')
    overlapping_blobs = [blob for blob in eroded_blobs_in_frame
                         if blob_to_test.overlaps_with(blob)]
    # logger.debug('Finished getting candidate blobs by overlapping')
    return overlapping_blobs if len(overlapping_blobs) > 0 else eroded_blobs_in_frame


def get_missing_identities_from_blobs_in_frame(possible_identities, blobs_in_frame, occluded_identities_in_frame):
    identities_in_frame = []
    for blob in blobs_in_frame:
        identities_in_frame.extend(blob.final_identities)
    return (set(possible_identities) - set(identities_in_frame)) - set(occluded_identities_in_frame)

def get_candidate_centroid(individual_gap_interval, previous_blob_to_the_gap, next_blob_to_the_gap, identity, border = '', inner_frame_number = None):
    # logger.debug('Getting candidate centroids')
    blobs_for_interpolation = [previous_blob_to_the_gap, next_blob_to_the_gap]
    centroids_to_interpolate = [blob_for_interpolation.final_centroids[blob_for_interpolation.final_identities.index(identity)]
                                for blob_for_interpolation in blobs_for_interpolation]
    centroids_to_interpolate = np.asarray(list(zip(*centroids_to_interpolate)))
    argsort_x = np.argsort(centroids_to_interpolate[0])
    centroids_to_interpolate[0] = centroids_to_interpolate[0][argsort_x]
    centroids_to_interpolate[1] = centroids_to_interpolate[1][argsort_x]
    number_of_points = individual_gap_interval[1] - individual_gap_interval[0] + 1
    x_interp = np.linspace(centroids_to_interpolate[0][0], centroids_to_interpolate[0][1], number_of_points + 1)
    y_interp = np.interp(x_interp, centroids_to_interpolate[0], centroids_to_interpolate[1])
    if border == 'start' and np.all(argsort_x == np.asarray([0, 1])):
        # logger.debug('Finished getting candidate centroids')
        return list(zip(x_interp, y_interp))[1]
    elif border == 'start' and np.all(argsort_x == np.asarray([1, 0])):
        # logger.debug('Finished getting candidate centroids')
        return list(zip(x_interp, y_interp))[-2]
    elif border == 'end' and np.all(argsort_x == np.asarray([0, 1])):
        # logger.debug('Finished getting candidate centroids')
        return list(zip(x_interp, y_interp))[-2]
    elif border == 'end' and np.all(argsort_x == np.asarray([1, 0])):
        # logger.debug('Finished getting candidate centroids')
        return list(zip(x_interp, y_interp))[1]
    else:
        raise ValueError('border must be start or end: %s was given instead' %border)

def find_the_individual_gap_interval(blobs_in_video, possible_identities, identity, a_frame_in_the_gap, list_of_occluded_identities):
    # logger.debug('Finding the individual gap interval')
    # find gap start
    identity_is_missing = True
    gap_start = a_frame_in_the_gap
    frame_number = gap_start

    # logger.debug('To the past while loop')
    while identity_is_missing and frame_number > 0 and frame_number < len(blobs_in_video):
        blobs_in_frame = blobs_in_video[frame_number]
        occluded_identities_in_frame = list_of_occluded_identities[frame_number]
        missing_identities = get_missing_identities_from_blobs_in_frame(possible_identities, blobs_in_frame, occluded_identities_in_frame)
        if identity not in missing_identities:
            gap_start = frame_number + 1
            identity_is_missing = False
        else:
            frame_number -= 1

    # find gap end
    identity_is_missing = True
    frame_number = a_frame_in_the_gap
    gap_end = a_frame_in_the_gap

    # logger.debug('To the future while loop')
    while identity_is_missing and frame_number > 0 and frame_number < len(blobs_in_video):
        blobs_in_frame = blobs_in_video[frame_number]
        occluded_identities_in_frame = list_of_occluded_identities[frame_number]
        missing_identities = get_missing_identities_from_blobs_in_frame(possible_identities, blobs_in_frame, occluded_identities_in_frame)
        if identity not in missing_identities:
            gap_end = frame_number
            identity_is_missing = False
        else:
            gap_end += 1
            frame_number = gap_end
    # logger.debug('Finished finding the individual gap interval')
    return (gap_start, gap_end)

def get_previous_and_next_blob_wrt_gap(blobs_in_video, possible_identities, identity, frame_number, list_of_occluded_identities):
    # logger.debug('Finding previons and next blobs to the gap of this identity')
    individual_gap_interval = find_the_individual_gap_interval(blobs_in_video,
                                        possible_identities,
                                        identity,
                                        frame_number,
                                        list_of_occluded_identities)
    if individual_gap_interval[0] != 0:
        previous_blob_to_the_gap = get_blob_by_identity(blobs_in_video[individual_gap_interval[0]-1], identity)
    else:
        previous_blob_to_the_gap = None
    if individual_gap_interval[1] != len(blobs_in_video):
        next_blob_to_the_gap = get_blob_by_identity(blobs_in_video[individual_gap_interval[1]], identity)
    else:
        next_blob_to_the_gap = None
    if previous_blob_to_the_gap is not None and len(previous_blob_to_the_gap) == 1 and previous_blob_to_the_gap[0] is not None:
        previous_blob_to_the_gap = previous_blob_to_the_gap[0]
    else:
        previous_blob_to_the_gap = None
    if  next_blob_to_the_gap is not None and len(next_blob_to_the_gap) == 1 and next_blob_to_the_gap[0] is not None:
        next_blob_to_the_gap = next_blob_to_the_gap[0]
    else:
        next_blob_to_the_gap = None
    # logger.debug('Finished finding previons and next blobs to the gap of this identity')
    return individual_gap_interval, previous_blob_to_the_gap, next_blob_to_the_gap


def get_closest_contour_point_to(contour, candidate_centroid):
    return tuple(contour[np.argmin(cdist([candidate_centroid], np.squeeze(contour)))][0])


def get_nearest_eroded_blob_to_candidate_centroid(eroded_blobs, candidate_centroid, identity=None, inner_frame_number=None):
    eroded_blob_index = np.argmin([blob.distance_from_countour_to(candidate_centroid) for blob in eroded_blobs])
    return eroded_blobs[eroded_blob_index]


def nearest_candidate_blob_is_near_enough(video, candidate_blob, candidate_centroid, blob_in_border_frame):
    points = [candidate_centroid, blob_in_border_frame.centroid]
    distances = np.asarray([np.sqrt(candidate_blob.squared_distance_to(point)) for point in points])
    return np.any(distances < video.velocity_threshold)


def eroded_blob_overlaps_with_blob_in_border_frame(eroded_blob, blob_in_border_frame):
    return eroded_blob.overlaps_with(blob_in_border_frame)


def centroid_is_inside_of_any_eroded_blob(candidate_eroded_blobs, candidate_centroid):
    # logger.debug('Checking whether the centroids is inside of a blob')
    candidate_centroid = tuple([int(centroid_coordinate) for centroid_coordinate in candidate_centroid])
    # logger.debug('Finished whether the centroids is inside of a blob')
    return [blob for blob in candidate_eroded_blobs
            if cv2.pointPolygonTest(blob.contour, candidate_centroid, False) >= 0]


def evaluate_candidate_blobs_and_centroid(video, candidate_eroded_blobs, candidate_centroid, blob_in_border_frame, blobs_in_frame = None, inner_frame_number = None, identity = None):
    # logger.debug('Evaluating candidate blobs and centroids')
    blob_containing_candidate_centroid = centroid_is_inside_of_any_eroded_blob(candidate_eroded_blobs, candidate_centroid)
    if blob_containing_candidate_centroid:
        # logger.debug('Finished evaluating candidate blobs and centroids: the candidate centroid is in an eroded blob')
        return blob_containing_candidate_centroid[0], candidate_centroid
    elif len(candidate_eroded_blobs) > 0:
        nearest_blob = get_nearest_eroded_blob_to_candidate_centroid(candidate_eroded_blobs, candidate_centroid, identity, inner_frame_number)
        new_centroid = get_closest_contour_point_to(nearest_blob.contour, candidate_centroid)
        if nearest_candidate_blob_is_near_enough(video, nearest_blob, candidate_centroid, blob_in_border_frame) or \
            eroded_blob_overlaps_with_blob_in_border_frame(nearest_blob, blob_in_border_frame):
            # logger.debug('Finished evaluating candidate blobs and centroids: the candidate centroid is near to a candidate blob')
            return nearest_blob, new_centroid
        else:
            # logger.debug('Finished evaluating candidate blobs and centroids: the candidate centrois is far from a candidate blob')
            return None, None
    else:
        # logger.debug('Finished evaluating candidate blobs and centroids: there where no candidate blobs')
        return None, None


def get_candidate_tuples_with_centroids_in_original_blob(original_blob, candidate_tuples_to_close_gap):
    candidate_tuples_with_centroids_in_original_blob = [candidate_tuple for candidate_tuple in candidate_tuples_to_close_gap
                                                        if cv2.pointPolygonTest(original_blob.contour,
                                                        tuple([int(c) for c in candidate_tuple[1]]), False) >= 0]
    return candidate_tuples_with_centroids_in_original_blob


def assign_identity_to_new_blobs(video, fragments, blobs_in_video, possible_identities, original_inner_blobs_in_frame, candidate_tuples_to_close_gap, list_of_occluded_identities):
    # logger.debug('Assigning identity to new blobs')
    new_original_blobs = []

    for i, original_blob in enumerate(original_inner_blobs_in_frame):
        # logger.debug('Checking original blob')
        candidate_tuples_with_centroids_in_original_blob = get_candidate_tuples_with_centroids_in_original_blob(original_blob, candidate_tuples_to_close_gap)
        if len(candidate_tuples_with_centroids_in_original_blob) == 1: # the gap is a single individual blob
            # logger.debug('Only a candidate tuple for this original blob')
            identity = candidate_tuples_with_centroids_in_original_blob[0][2]
            centroid = candidate_tuples_with_centroids_in_original_blob[0][1]
            if original_blob.is_an_individual and len(original_blob.final_identities) == 1 and original_blob.final_identities[0] == 0:
                original_blob._identities_corrected_closing_gaps = [identity]
                [setattr(blob,'_identities_corrected_closing_gaps', [identity])
                 for blobs_in_frame in blobs_in_video for blob in blobs_in_frame
                 if blob.fragment_identifier == original_blob.fragment_identifier]
            elif original_blob.is_an_individual:
                list_of_occluded_identities[original_blob.frame_number].append(identity)
            elif original_blob.is_a_crossing:
                if original_blob.identities_corrected_closing_gaps is not None:
                    identity = original_blob.identities_corrected_closing_gaps + [identity]
                    centroid = original_blob.interpolated_centroids + [centroid]
                else:
                    identity = [identity]
                    centroid = [centroid]
                frame_number = original_blob.frame_number
                new_blob = candidate_tuples_with_centroids_in_original_blob[0][0]
                new_blob.frame_number = frame_number
                new_blob._identities_corrected_closing_gaps = identity
                new_blob.interpolated_centroids = centroid
                original_blob = new_blob

            new_original_blobs.append(original_blob)
        elif len(candidate_tuples_with_centroids_in_original_blob) > 1 and original_blob.is_a_crossing: # Note that the original blobs that were unidentified (identity 0) are set to zero before starting the main while loop
            # logger.debug('Many candidate tuples for this original blob, and the original blob is a crossing')
            candidate_eroded_blobs = list(zip(*candidate_tuples_with_centroids_in_original_blob))[0]
            candidate_eroded_blobs_centroids = list(zip(*candidate_tuples_with_centroids_in_original_blob))[1]
            candidate_eroded_blobs_identities = list(zip(*candidate_tuples_with_centroids_in_original_blob))[2]
            if len(set(candidate_eroded_blobs)) == 1: # crossing not split
                original_blob.interpolated_centroids = [candidate_eroded_blob_centroid
                                                                    for candidate_eroded_blob_centroid in candidate_eroded_blobs_centroids]
                original_blob._identities_corrected_closing_gaps = [candidate_eroded_blob_identity
                                                                    for candidate_eroded_blob_identity in candidate_eroded_blobs_identities]
                original_blob.eroded_pixels = candidate_eroded_blobs[0].pixels
                new_original_blobs.append(original_blob)

            elif len(set(candidate_eroded_blobs)) > 1: # crossing split
                list_of_new_blobs_in_next_frames = []
                count_eroded_blobs = {eroded_blob: candidate_eroded_blobs.count(eroded_blob)
                                        for eroded_blob in candidate_eroded_blobs}
                for j, (eroded_blob, centroid, identity) in enumerate(candidate_tuples_with_centroids_in_original_blob):
                    if count_eroded_blobs[eroded_blob] == 1: # split blob, single individual
                        eroded_blob.frame_number = original_blob.frame_number
                        eroded_blob.centroid = centroid
                        eroded_blob._identities_corrected_closing_gaps = [identity]
                        eroded_blob._is_an_individual = True
                        eroded_blob._was_a_crossing = True
                        new_original_blobs.append(eroded_blob)
                    elif count_eroded_blobs[eroded_blob] > 1:
                        if not hasattr(eroded_blob, 'interpolated_centroids'):
                            eroded_blob.interpolated_centroids = []
                            eroded_blob._identities_corrected_closing_gaps = []
                        eroded_blob.frame_number = original_blob.frame_number
                        eroded_blob.interpolated_centroids.append(centroid)
                        eroded_blob._identities_corrected_closing_gaps.append(identity)
                        eroded_blob._is_a_crossing = True
                        new_original_blobs.append(eroded_blob)

        new_original_blobs.append(original_blob)

    new_original_blobs = list(set(new_original_blobs))
    blobs_in_video[original_blob.frame_number] = new_original_blobs
    # logger.debug('Finished assigning identity to new blobs')
    return blobs_in_video, list_of_occluded_identities


def get_forward_backward_list_of_frames(gap_interval):
    """input:
    gap_interval: array of tuple [start_frame_number, end_frame_number]
    output:
    [f1, fn, f2, fn-1, ...] for f1 = start_frame_number and
                                fn = end_frame_number"""
    # logger.debug('Got forward-backward list of frames')
    gap_range = range(gap_interval[0],gap_interval[1])
    gap_length = len(gap_range)
    return np.insert(gap_range[::-1], np.arange(gap_length), gap_range)[:gap_length]


def interpolate_trajectories_during_gaps(video, list_of_blobs, list_of_fragments, list_of_occluded_identities, possible_identities, erosion_counter):
    # logger.debug('In interpolate_trajectories_during_gaps')
    blobs_in_video = list_of_blobs.blobs_in_video
    for frame_number, (blobs_in_frame, occluded_identities_in_frame) in enumerate(tqdm(zip(blobs_in_video, list_of_occluded_identities), desc = "closing gaps")):
        if frame_number != 0:
            # logger.debug('-Main frame number %i' %frame_number)
            # logger.debug('Getting missing identities')
            missing_identities = get_missing_identities_from_blobs_in_frame(possible_identities, blobs_in_frame, occluded_identities_in_frame)
            if len(missing_identities) > 0 and len(blobs_in_frame) >= 1:
                gap_interval = find_the_gap_interval(blobs_in_video, possible_identities, frame_number, list_of_occluded_identities)
                forward_backward_list_of_frames = get_forward_backward_list_of_frames(gap_interval)
                # logger.debug('--There are missing identities in this main frame: gap interval %s ' %(gap_interval,))
                for index, inner_frame_number in enumerate(forward_backward_list_of_frames):
                    # logger.debug('---Length forward_backward_list_of_frames %i' %len(forward_backward_list_of_frames))
                    # logger.debug('---Gap interval: interval %s ' %(gap_interval,))
                    # logger.debug('---Inner frame number %i' %inner_frame_number )
                    inner_occluded_identities_in_frame = list_of_occluded_identities[inner_frame_number]
                    inner_blobs_in_frame = blobs_in_video[inner_frame_number]
                    if len(inner_blobs_in_frame) != 0:
                        # logger.debug('----There are blobs in the inner frame')
                        if erosion_counter != 0:
                            eroded_blobs_in_frame = get_eroded_blobs(video, inner_blobs_in_frame, inner_frame_number) #list of eroded blobs!
                            if len(eroded_blobs_in_frame) == 0:
                                eroded_blobs_in_frame = inner_blobs_in_frame
                        else:
                            eroded_blobs_in_frame = inner_blobs_in_frame
                        # logger.debug('Getting missing identities')
                        inner_missing_identities = get_missing_identities_from_blobs_in_frame(possible_identities,
                                                                                            inner_blobs_in_frame,
                                                                                            inner_occluded_identities_in_frame)
                        candidate_tuples_to_close_gap = []
                        for identity in inner_missing_identities:
                            # logger.debug('-----Solving identity %i' %identity)
                            individual_gap_interval,\
                            previous_blob_to_the_gap,\
                            next_blob_to_the_gap = get_previous_and_next_blob_wrt_gap(blobs_in_video,
                                                                                        possible_identities,
                                                                                        identity,
                                                                                        inner_frame_number,
                                                                                        list_of_occluded_identities)
                            # logger.debug('individual_gap_interval: %s' %(individual_gap_interval,))
                            if previous_blob_to_the_gap is not None and next_blob_to_the_gap is not None:
                                # logger.debug('------The previous and next blobs are not None')
                                border = 'start' if index % 2 == 0 else 'end'
                                candidate_centroid = get_candidate_centroid(individual_gap_interval,
                                                                            previous_blob_to_the_gap,
                                                                            next_blob_to_the_gap,
                                                                            identity,
                                                                            border = border,
                                                                            inner_frame_number = inner_frame_number)
                                if border == 'start':
                                    blob_in_border_frame = previous_blob_to_the_gap
                                elif border == 'end':
                                    blob_in_border_frame = next_blob_to_the_gap
                                candidate_eroded_blobs_by_overlapping = get_candidate_blobs_by_overlapping(blob_in_border_frame,
                                                                                                    eroded_blobs_in_frame)
                                candidate_eroded_blobs_by_inclusion_of_centroid = centroid_is_inside_of_any_eroded_blob(eroded_blobs_in_frame, candidate_centroid)
                                candidate_eroded_blobs = candidate_eroded_blobs_by_overlapping + candidate_eroded_blobs_by_inclusion_of_centroid
                                candidate_blob_to_close_gap, centroid = evaluate_candidate_blobs_and_centroid(video,
                                                                                                                candidate_eroded_blobs,
                                                                                                                candidate_centroid,
                                                                                                                blob_in_border_frame,
                                                                                                                blobs_in_frame = inner_blobs_in_frame,
                                                                                                                inner_frame_number = inner_frame_number, identity = identity)
                                if candidate_blob_to_close_gap is not None:
                                    # logger.debug('------There is a tuple (blob, centroid, identity) to close the gap in this inner frame)')
                                    candidate_tuples_to_close_gap.append((candidate_blob_to_close_gap, centroid, identity))
                                else:
                                    # logger.debug('------There are no candidate blobs and/or centroids: it must be occluded or it jumped')
                                    list_of_occluded_identities[inner_frame_number].append(identity)
                            else: # this manages the case in which identities are missing in the first frame or disappear without appearing anymore,
                                # and evntual occlusions (an identified blob does not appear in the previous and/or the next frame)
                                # logger.debug('------There is not next or not previous blob to this inner gap: it must be occluded')
                                # logger.debug('previous_blob_to_the_gap is None') if previous_blob_to_the_gap is None else logger.debug('previous_blob_to_the_gap exists')
                                # logger.debug('next_blob_to_the_gap is None') if next_blob_to_the_gap is None else logger.debug('next_blob_to_the_gap exists')
                                [list_of_occluded_identities[i].append(identity) for i in range(individual_gap_interval[0],individual_gap_interval[1])]

                        # logger.debug('-----Assinning identities to candidate tuples (blob, centroid, identity)')
                        blobs_in_video, list_of_occluded_identities = assign_identity_to_new_blobs(video, list_of_fragments.fragments,
                                                                    blobs_in_video, possible_identities,
                                                                    inner_blobs_in_frame, candidate_tuples_to_close_gap,
                                                                    list_of_occluded_identities)
                # else:
                    # logger.debug('----No blobs in this inner frame')
            # else:
                # logger.debug('--No missing identities in this frame')
        # else:
            # logger.debug('-We do not check the first frame')
    return blobs_in_video, list_of_occluded_identities


def get_number_of_non_split_crossing(blobs_in_video):
    return len([blob for blobs_in_frame in blobs_in_video
                for blob in blobs_in_frame
                if blob.is_a_crossing])


def reset_blobs_in_video_before_erosion_iteration(blobs_in_video):
    """Resets the identity of crossings and individual with multiple identities
    before starting a loop of interpolation

    Parameters
    ----------
    blobs_in_video : list of lists of `Blob` objects
    """
    # logger.debug('Reseting blobs to start erosion iteration')
    for blobs_in_frame in blobs_in_video:
        for blob in blobs_in_frame:
            if blob.is_a_crossing:
                blob._identity = None
            elif blob.is_an_individual and len(blob.final_identities) > 1:
                blob._identities_corrected_closing_gaps = None


def closing_gap_stopping_criteria(blobs_in_video, previous_number_of_non_split_crossings_blobs):
    current_number_of_non_split_crossings = get_number_of_non_split_crossing(blobs_in_video)
    return current_number_of_non_split_crossings, previous_number_of_non_split_crossings_blobs > current_number_of_non_split_crossings


def clean_individual_blob_before_saving(blobs_in_video):
    """Clean inidividual blobs whose identity is a list (it cannot be, hence an
    occluded identity has been assigned to an individual blob).
    """
    for blobs_in_frame in blobs_in_video:
        for blob in blobs_in_frame:
            if blob.is_an_individual and len(blob.final_identities) > 1:
                blob._identities_corrected_closing_gaps = None

    return blobs_in_video

def close_trajectories_gaps(video, list_of_blobs, list_of_fragments):
    """This is the main function to close the gaps where animals have not been
    identified (labelled with identity 0), are crossing with another animals or
    are occluded or not segmented.

    Parameters
    ----------
    video : <Video object>
        Object containing all the parameters of the video.
    list_of_blobs : <ListOfBlobs object>
        Object with the collection of blobs found during segmentation with associated
        methods. See :class:`list_of_blobs.ListOfBlobs`
    list_of_fragments : <ListOfFragments object>
        Collection of individual and crossing fragments with associated methods.
        See :class:`list_of_fragments.ListOfFragments`

    Returns
    -------
    list_of_blobs : <ListOfBlobs object>
        ListOfBlobs object with the updated blobs and identities that close gaps

    See Also
    --------
    :func:`set_individual_with_identity_0_as_crossings`
    :func:`compute_erosion_disk`
    :func:`compute_model_velocity`
    :func:`reset_blobs_in_video_before_erosion_iteration`
    :func:`interpolate_trajectories_during_gaps`
    :func:`closing_gap_stopping_criteria`
    :func:`clean_individual_blob_before_saving`

    """
    logger.debug('********************************')
    logger.debug('Starting close_trajectories_gaps')
    set_individual_with_identity_0_as_crossings(list_of_blobs)
    continue_erosion_protocol = True
    previous_number_of_non_split_crossings_blobs = sum([fragment.number_of_images
                                                        for fragment in list_of_fragments.fragments
                                                        if fragment.is_a_crossing])
    if not hasattr(video, '_erosion_kernel_size'):
        video._erosion_kernel_size = compute_erosion_disk(video, list_of_blobs.blobs_in_video)
        video.save()
    if not hasattr(video, 'velocity_threshold'):
        video.velocity_threshold = compute_model_velocity(list_of_fragments.fragments,
                                                          video.number_of_animals,
                                                          percentile = conf.VEL_PERCENTILE)
    possible_identities = range(1, video.number_of_animals + 1)
    erosion_counter = 0
    list_of_occluded_identities = [[] for i in range(len(list_of_blobs.blobs_in_video))]

    while continue_erosion_protocol or erosion_counter == 1:
        reset_blobs_in_video_before_erosion_iteration(list_of_blobs.blobs_in_video)
        list_of_blobs.blobs_in_video, \
        list_of_occluded_identities = \
            interpolate_trajectories_during_gaps(video,
                                                 list_of_blobs,
                                                 list_of_fragments,
                                                 list_of_occluded_identities,
                                                 possible_identities,
                                                 erosion_counter)
        current_number_of_non_split_crossings, continue_erosion_protocol = closing_gap_stopping_criteria(list_of_blobs.blobs_in_video,
                                                                                previous_number_of_non_split_crossings_blobs)
        previous_number_of_non_split_crossings_blobs = current_number_of_non_split_crossings
        erosion_counter += 1

    if not video.is_centroid_updated:
        list_of_blobs.blobs_in_video = clean_individual_blob_before_saving(list_of_blobs.blobs_in_video)
    return list_of_blobs



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

from __future__ import absolute_import, print_function, division
import numpy as np
import cv2
from tqdm import tqdm
from idtrackerai.utils.py_utils import  get_spaced_colors_util

def init(video_path, trajectories_dict_path):
    video_object = np.load(video_path, allow_pickle=True, encoding='latin1').item()
    video_object.update_paths(video_path)
    trajectories  = np.load(trajectories_dict_path, allow_pickle=True, encoding='latin1').item()['trajectories']
    colors = get_spaced_colors_util(video_object.number_of_animals, black = False)
    path_to_save_video = video_object._session_folder +'/tracked.avi'
    fourcc = cv2.VideoWriter_fourcc(*'XVID')
    video_writer = cv2.VideoWriter(path_to_save_video, fourcc, video_object.frames_per_second,
                                    (video_object.width, video_object.height))
    print("The video will be saved at ", path_to_save_video)
    return video_object, trajectories, video_writer, colors

def generate_video(video_object,
                trajectories,
                video_writer,
                colors,
                func = None,
                centroid_trace_length = 10,
                starting_frame = 0,
                ending_frame = None):
    if video_object.paths_to_video_segments is None:
        cap = cv2.VideoCapture(video_object.video_path)
    else:
        cap = None

    if ending_frame is None:
        ending_frame = len(trajectories)


    for frame_number in tqdm(range(starting_frame, ending_frame), desc='Adding trajectories to video'):
        frame = apply_func_on_frame(video_object,
                            trajectories,
                            frame_number,
                            video_writer,
                            colors,
                            cap = cap,
                            func = writeIds,
                            centroid_trace_length  = centroid_trace_length)
        video_writer.write(frame)

def apply_func_on_frame(video_object,
                        trajectories,
                        frame_number,
                        video_writer,
                        colors,
                        cap = None,
                        func = None,
                        centroid_trace_length = 10):
    segment_number = video_object.in_which_episode(frame_number)
    current_segment = segment_number
    if cap is None:
        cap = cv2.VideoCapture(video_object.paths_to_video_segments[segment_number])
        start = video_object._episodes_start_end[segment_number][0]
        cap.set(1,frame_number - start)
    else:
        cap.set(1, frame_number)
    ret, frame = cap.read()
    if ret:
        if video_object.resolution_reduction != 1:
            frame = cv2.resize(frame, None,
                                fx = video_object.resolution_reduction,
                                fy = video_object.resolution_reduction,
                                interpolation = cv2.INTER_AREA)
        frame = func(video_object, frame, frame_number, trajectories, centroid_trace_length,
                    colors)
        return frame


def writeIds(video_object, frame, frame_number, trajectories, centroid_trace_length, colors):
    ordered_centroid = trajectories[frame_number]
    font = cv2.FONT_HERSHEY_SIMPLEX
    font_size = 1 * video_object.resolution_reduction
    font_width = int(3 * video_object.resolution_reduction)
    font_width = 1 if font_width == 0 else font_width
    circle_size = int(2 * video_object.resolution_reduction)

    for cur_id, centroid in enumerate(ordered_centroid):
        if sum(np.isnan(centroid)) == 0:
            if frame_number > centroid_trace_length:
                centroids_trace = trajectories[frame_number - centroid_trace_length : frame_number, cur_id]
            else:
                centroids_trace = trajectories[: frame_number, cur_id]
            cur_id_str = str(cur_id + 1)
            int_centroid = np.asarray(centroid).astype('int')
            cv2.circle(frame, tuple(int_centroid), circle_size, colors[cur_id], -1)
            cv2.putText(frame, cur_id_str,tuple(int_centroid), font, font_size, colors[cur_id], font_width)
            for centroid_trace in centroids_trace:
                if sum(np.isnan(centroid_trace)) == 0:
                    int_centroid = np.asarray(centroid_trace).astype('int')
                    cv2.circle(frame, tuple(int_centroid), circle_size, colors[cur_id], -1)
    return frame

def main(args):
    video_object_path = args.video_object_path
    trajectories_dict_path = args.trajectories_path
    number_of_points_in_past_trace = args.number_of_ghost_points
    video_object, trajectories, video_writer, colors = init(video_object_path,
                                                    trajectories_dict_path)
    generate_video(video_object,
                    trajectories,
                    video_writer,
                    colors,
                    func = writeIds,
                    centroid_trace_length = number_of_points_in_past_trace,
                    starting_frame = args.starting_frame,
                    ending_frame = args.ending_frame)

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--video_object_path", type = str, help = "Path to the video object created during the tracking session")
    parser.add_argument("-t", "--trajectories_path", type = str, help = "Path to the trajectory file")
    parser.add_argument("-s", "--number_of_ghost_points", type = int, default = 20,
                        help = "Number of points used to draw the individual trajectories' traces")
    parser.add_argument("-sf", "--starting_frame", type = int, default = 0,
                        help = "Frame where to start the video")
    parser.add_argument("-ef", "--ending_frame", type = int, default = None,
                        help = "Frame where to end the video")
    args = parser.parse_args()
    main(args)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys
import numpy as np

from idtrackerai.network.identification_model.get_data import DataSet
from idtrackerai.network.identification_model.get_predictions import GetPrediction
from idtrackerai.network.identification_model.id_CNN import ConvNetwork # for documentation
from idtrackerai.list_of_fragments import ListOfFragments # for documentation

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.assigner")

"""
Identification of individual fragments given the predictions generate by the idCNN
"""

def assign(net, images):
    """Gathers the predictions relative to the images contained in `images`.
    Such predictions are returned as attributes of `assigner`.

    Parameters
    ----------
    net : <ConvNetwork object>
        Convolutional neural network object created according to net.params
    images : ndarray
        array of images

    Returns
    -------
    <GetPrediction object>
        The assigner object has as main attributes the list of predictions
        associated to `images` and the the corresponding softmax vectors

    See Also
    --------
    GetPrediction
    """
    logger.info("assigning identities to images...")
    images = np.expand_dims(np.asarray(images), axis=3)
    logger.info("generating data set. Images shape %s" %str(images.shape))
    data = DataSet(net.params.number_of_animals, images)
    logger.info("getting predictions")
    assigner = GetPrediction(data)
    assigner.get_predictions_softmax(net.predict)
    logger.info("done")
    return assigner


def compute_identification_statistics_for_non_accumulated_fragments(fragments, assigner, number_of_animals=None):
    """Given the predictions associated to the images in each (individual)
    fragment in the list fragments if computes the statistics necessary for the
    identification of fragment.

    Parameters
    ----------
    fragments : list
        List of individual fragment objects
    assigner : <GetPrediction object>
        The assigner object has as main attributes the list of predictions
        associated to `images` and the the corresponding softmax vectors
    number_of_animals : int
        number of animals to be tracked
    """
    counter = 0
    for fragment in fragments:
        if not fragment.used_for_training and fragment.is_an_individual:
            next_counter_value = counter + fragment.number_of_images
            predictions = assigner.predictions[counter : next_counter_value]
            softmax_probs = assigner.softmax_probs[counter : next_counter_value]
            fragment.compute_identification_statistics(predictions, softmax_probs, number_of_animals=number_of_animals)
            counter = next_counter_value

def assign_identity(list_of_fragments):
    """Identifies the individual fragments recursively, based on the value of
    P2

    Parameters
    ----------
    list_of_fragments : <ListOfFragments object>
        collection of the individual fragments and associated methods
    """
    list_of_fragments.compute_P2_vectors()
    number_of_unidentified_individual_fragments = list_of_fragments.get_number_of_unidentified_individual_fragments()

    while number_of_unidentified_individual_fragments != 0:
        fragment = list_of_fragments.get_next_fragment_to_identify()
        fragment.assign_identity()
        number_of_unidentified_individual_fragments -= 1


def assign_remaining_fragments(list_of_fragments, video, net):
    """This is the main function of this module: given a list_of_fragments it
    puts in place the routine to identify, if possible, each of the individual
    fragments. The starting point for the identification is given by the
    predictions produced by the ConvNetwork net passed as input. The organisation
    of the images in individual fragments is then used to assign more accurately.

    Parameters
    ----------
    list_of_fragments : <ListOfFragments object>
        collection of the individual fragments and associated methods
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    net : <ConvNetwork object>
        Convolutional neural network object created according to net.params

    See Also
    --------
    ListOfFragments.get_images_from_fragments_to_assign
    assign
    compute_identification_statistics_for_non_accumulated_fragments

    """
    logger.info("Assigning identities to non-accumulated individual fragments")
    logger.debug("Resetting list of fragments for assignment")
    list_of_fragments.reset(roll_back_to = 'accumulation')
    number_of_unidentified_individual_fragments = list_of_fragments.get_number_of_unidentified_individual_fragments()
    logger.info("Number of unidentifier individual fragments: {}".format(number_of_unidentified_individual_fragments))
    if number_of_unidentified_individual_fragments != 0:
        logger.info("Getting images")
        images = list_of_fragments.get_images_from_fragments_to_assign()
        logger.debug("Images shape before assignment %s" %str(images.shape))
        logger.info("Getting predictions")
        assigner = assign(net, images)
        logger.debug("Number of generated predictions: %s" %str(len(assigner._predictions)))
        logger.debug("Predictions range: %s" %str(np.unique(assigner._predictions)))
        compute_identification_statistics_for_non_accumulated_fragments(list_of_fragments.fragments, assigner)
        logger.info("Assigning identities")
        assign_identity(list_of_fragments)
    else:
        list_of_fragments.compute_P2_vectors()



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import os
import sys

import matplotlib.pyplot as plt
import numpy as np
from confapp import conf

from idtrackerai.network.identification_model.get_data import split_data_train_and_validation
from idtrackerai.network.identification_model.epoch_runner import EpochRunner
from idtrackerai.network.identification_model.stop_training_criteria import Stop_Training
from idtrackerai.network.identification_model.store_accuracy_and_loss import Store_Accuracy_and_Loss

if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.trainer")


def train(video,
          fragments,
          net,
          images,
          labels,
          store_accuracy_and_error,
          check_for_loss_plateau,
          save_summaries,
          print_flag,
          plot_flag,
          global_step=0,
          accumulation_manager=None,
          batch_size=conf.BATCH_SIZE_IDCNN):
    """Short summary.

    Parameters
    ----------
    video : <Video object>
        an instance of the class :class:`~video.Video`
    fragments : list
        list of instances of the class :class:`~fragment.Fragment`
    net : <ConvNetwork object>
        an instance of the class :class:`~id_CNN.ConvNetwork`
    images : ndarray
        array of shape [number_of_images, height, width]
    labels : type
        array of shape [number_of_images, number_of_animals]
    store_accuracy_and_error : bool
        if True the values of the loss function, accuracy and individual
        accuracy will be stored
    check_for_loss_plateau : bool
        if True the stopping criteria (see :mod:`~stop_training_criteria`) will
        automatically stop the training in case the loss functin computed for
        the validation set of images reaches a plateau
    sasave_summaries : bool
        if True tensorflow summaries will be generated and stored to allow
        tensorboard visualisation of both loss and activity histograms
    print_flag : bool
        if True additional information are printed in the terminal
    plot_flag : bool
        if True training and validation loss, accuracy and individual accuracy
        are plot in a graph at the end of the training session
    global_epoch : int
        global counter of the training epoch in pretraining
    accumulation_manager : <AccumulationManager object>
        an instance of the class
        :class:`~accumulation_manager.AccumulationManager`
    batch_size : int
        size of the batch of images used for training

    Returns
    -------
    int
        global epoch counter updated after the training session
    <ConvNetwork object>
        network with updated parameters after training
    float
        ration of images used for pretraining over the total number of
        available images
    <Store_Accuracy_and_Loss object>
        updated with the values collected on the training set of labelled
        images
    <Store_Accuracy_and_Loss object>
        updated with the values collected on the validation set of labelled
        images
    """
    # Save accuracy and error during training and validation
    # The loss and accuracy of the validation are saved to allow the automatic stopping of the training
    logger.info("Training...")
    store_training_accuracy_and_loss_data = Store_Accuracy_and_Loss(net, name = 'training', scope = 'training')
    store_validation_accuracy_and_loss_data = Store_Accuracy_and_Loss(net, name = 'validation', scope = 'training')
    if plot_flag:
        plt.ion()
        fig, ax_arr = plt.subplots(4)
        title = 'Accumulation-' + str(video.accumulation_trial) + '-' + str(video.accumulation_step)
        fig.canvas.set_window_title(title)
        fig.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.5)
    # Instantiate data set
    training_dataset, validation_dataset = split_data_train_and_validation(net.params.number_of_animals, images, labels)
    # Convert labels to one hot vectors
    training_dataset.convert_labels_to_one_hot()
    validation_dataset.convert_labels_to_one_hot()
    # Reinitialize softmax and fully connected
    if video is None or video.accumulation_step == 0:
        net.reinitialize_softmax_and_fully_connected()
    # Train network
    # compute weights to be fed to the loss function (weighted cross entropy)
    net.compute_loss_weights(training_dataset.labels)
    trainer = EpochRunner(training_dataset,
                        starting_epoch = global_step,
                        print_flag = print_flag,
                        batch_size = batch_size)
    validator = EpochRunner(validation_dataset,
                        starting_epoch = global_step,
                        print_flag = print_flag,
                        batch_size = batch_size)
    #set criteria to stop the training
    stop_training = Stop_Training(net.params.number_of_animals,
                                check_for_loss_plateau = check_for_loss_plateau,
                                first_accumulation_flag = video is None or video.accumulation_step == 0)

    global_step0 = global_step

    while not stop_training(store_training_accuracy_and_loss_data,
                            store_validation_accuracy_and_loss_data,
                            trainer._epochs_completed):
        # --- Training
        feed_dict_train = trainer.run_epoch('Training', store_training_accuracy_and_loss_data, net.train)
        # --- Validation
        feed_dict_val = validator.run_epoch('Validation', store_validation_accuracy_and_loss_data, net.validate)
        # update global step
        net.session.run(net.global_step.assign(trainer.starting_epoch + trainer._epochs_completed))
        # write summaries if asked
        if save_summaries:
            net.write_summaries(trainer.starting_epoch + trainer._epochs_completed, feed_dict_train, feed_dict_val)
        # Update counter
        trainer._epochs_completed += 1
        validator._epochs_completed += 1

    if (np.isnan(store_training_accuracy_and_loss_data.loss[-1]) or np.isnan(store_validation_accuracy_and_loss_data.loss[-1])):
        raise ValueError("The model diverged")
    else:
        global_step += trainer.epochs_completed
        #logger.debug('loss values in validation: %s' %str(store_validation_accuracy_and_loss_data.loss[global_step0:]))
        # update used_for_training flag to True for fragments used
        logger.info("Accumulation step completed. Updating global fragments used for training")
        if accumulation_manager is not None:
            accumulation_manager.update_fragments_used_for_training()
        # plot if asked
        if plot_flag:
            store_training_accuracy_and_loss_data.plot_global_fragments(ax_arr, video, fragments, black = False)
            store_training_accuracy_and_loss_data.plot(ax_arr, color='r')
            store_validation_accuracy_and_loss_data.plot(ax_arr, color='b')
            fig.savefig(os.path.join(net.params.save_folder, title + '.pdf'))
        # store training and validation losses and accuracies
        if store_accuracy_and_error:
            store_training_accuracy_and_loss_data.save(trainer._epochs_completed)
            store_validation_accuracy_and_loss_data.save(trainer._epochs_completed)
        net.save()
        return global_step, net, store_validation_accuracy_and_loss_data, store_training_accuracy_and_loss_data



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)
import sys
from setuptools import find_packages, setup
import re

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 6)
if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write("""
==========================
Unsupported Python version
==========================
This version of idtrackerai requires Python {}.{}, but you're trying to
install it on Python {}.{}.
""".format(*(REQUIRED_PYTHON + CURRENT_PYTHON)))
    sys.exit(1)

requirements = ['numpy >= 1.13.0',
                'natsort >= 5.0.2',
                'matplotlib >= 2.1',
                'seaborn >= 0.8',
                'tqdm >= 4.19',
                'joblib >= 0.11',
                'scikit-learn >= 0.19',
                'pyyaml >= 3.12',
                'psutil >= 5.4.3',
                'h5py >= 2.7.0',
                'xlib == 0.21',
                'msgpack-numpy >= 0.3.9',
                'msgpack-python >= 0.4.8',
                'pandas >= 0.20.2',
                'scipy >= 0.19.0',
                'sklearn >= 0.0',
                'tables >= 3.3.0',
                'dask >= 0.17.0',
                'opencv-python == 3.4.5.20',
                'tensorflow == 1.13.1',
                'idtrackerai-app',
                'confapp',
                'gdown'
                ]


EXCLUDE_FROM_PACKAGES = [ "plots", "plots.*", "docs", "docs.*"]

version = ''
with open('idtrackerai/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)

with open('README.md', 'r') as fd:
    long_description = fd.read()

setup(
    name='idtrackerai',
    version=version,
    python_requires='>={}.{}'.format(*REQUIRED_PYTHON),
    url='https://idtrackerai.readthedocs.io/en/latest/',
    author='Francisco Romero Ferrero, Mattia G. Bergomi, Francisco J.H. Heras, Ricardo Ribeiro',
    author_email='idtracker@gmail.com',
    description=('A multi-animal tracking algorithm based on convolutional neural networks'),
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    license='GPLv3+',
    packages=find_packages(exclude=EXCLUDE_FROM_PACKAGES),
    include_package_data=True,
    install_requires=requirements,
    extras_require={
        'gui': ['pyforms-gui'],
        'gpu': ['tensorflow-gpu==1.13.1']
    },
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    entry_points={
        'console_scripts': [
            'idtrackerai_test=idtrackerai.utils.test_example_video:test',
        ],
    }
)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import numpy as np
from tqdm import tqdm

"""
The correct_impossible_velocity_jumps module
"""

def reassign(fragment, fragments, impossible_velocity_threshold):
    """Reassigns the identity of a given `fragment` considering the identity of the
    `fragments` coexisting with it and the `impossible_velocity_threshold`

    Parameters
    ----------
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of overlapping
        blobs that are considered to be the same animal
    fragments : list
        List with all the `Fragment` objects of the video
    impossible_velocity_threshold : float
        If the velocity needed to link two fragments is higher than this threshold
        the identiy of one of the fragments is considerd to be wrong as it would be
        physically impossible for an animal to move so much. See `video.velocity_threshold`
        for each definition

    See Also
    --------
    :class:`fragment.Fragment`
    :meth:`get_available_and_non_available_identities`
    :meth:`get_candidate_identities_by_minimum_speed`
    :meth:`get_candidate_identities_above_random_P2`

    """
    def get_available_and_non_available_identities(fragment):
        """Computes the available and non available identities of a given fragment
        taking into consideration the identities of the fragments that coexist with it

        Parameters
        ----------
        fragment : <Fragment object>
            Object collecting all the information for a consecutive set of overlapping
            blobs that are considered to be the same animal

        Returns
        -------
        non_available_identities : nd.array
            Array with the non available identities for the `fragment`
        available_identities : set
            Set with the available idenities for the `fragment`
        See Also
        --------
        Fragment

        """
        non_available_identities = set([coexisting_fragment.assigned_identities[0]
                            for coexisting_fragment in fragment.coexisting_individual_fragments])
        available_identities = set(range(1, fragment.number_of_animals + 1)) - \
                            non_available_identities
        if fragment.assigned_identities[0] is not None and fragment.assigned_identities[0] != 0:
            available_identities = available_identities | set([fragment.assigned_identities[0]])
        if 0 in non_available_identities: non_available_identities.remove(0)
        non_available_identities = np.array(list(non_available_identities))
        return non_available_identities, available_identities

    def get_candidate_identities_by_minimum_speed(fragment,
                                                  fragments,
                                                  available_identities,
                                                  impossible_velocity_threshold):
        """Computes the candidate identities for a given `fragment` taking into
        consideration the velocities needed to join the `fragment` with its neighbour
        fragments in the past and in the future

        Parameters
        ----------
        fragment : <Fragment object>
            Object collecting all the information for a consecutive set of overlapping
            blobs that are considered to be the same animal
        fragments : list
            List with all the `Fragment` objects of the video
        available_identities : set
            Set with the available idenities for the `fragment`
        impossible_velocity_threshold : float
            If the velocity needed to link two fragments is higher than this threshold
            the identiy of one of the fragments is considerd to be wrong as it would be
            physically impossible for an animal to move so much. See `video.velocity_threshold`
            for each definition

        Returns
        -------
        candidate_identities_by_speed : nd.array
            Array with the identities that fullfill the `impossible_velocity_threshold`
            ordered from minimum to maximum velocity

        speed_of_candidate_identities : nd.array
            Array with the maximum velocity needed to link the given `fragment`
            with its neighbours assuming a given identity. Ordered from minimum to maximum
            velocity

        See Also
        --------
        Fragment
        compute_velocities_consecutive_fragments

        """
        speed_of_candidate_identities = []
        for identity in available_identities:
            fragment._user_generated_identity = identity
            neighbour_fragment_past = fragment.get_neighbour_fragment(fragments,
                                                                        'to_the_past')
            neighbour_fragment_future = fragment.get_neighbour_fragment(fragments,
                                                                        'to_the_future')
            velocities_between_fragments = compute_velocities_consecutive_fragments(neighbour_fragment_past,
                                                                        fragment,
                                                                        neighbour_fragment_future)

            if np.all(np.isnan(velocities_between_fragments)):
                speed_of_candidate_identities.append(impossible_velocity_threshold)
            else:
                speed_of_candidate_identities.append(np.nanmax(velocities_between_fragments))
        fragment._user_generated_identity = None
        argsort_identities_by_speed = np.argsort(speed_of_candidate_identities)
        return np.asarray(list(available_identities))[argsort_identities_by_speed], np.asarray(speed_of_candidate_identities)[argsort_identities_by_speed]

    def get_candidate_identities_above_random_P2(fragment, fragments,
                                                 non_available_identities,
                                                 available_identities,
                                                 impossible_velocity_threshold):
        """Computes the candidate identities of a `fragment` taking into consideration
        the probability of identification given by its `fragment.P2_vector`. An identity
        is a potential candidate if the probability of identification is above random.

        Parameters
        ----------
        fragment : <Fragment object>
            Object collecting all the information for a consecutive set of overlapping
            blobs that are considered to be the same animal
        fragments : list
            List with all the `Fragment` objects of the video
        non_available_identities : nd.array
            Array with the non available identities for the `fragment`
        available_identities : set
            Set with the available idenities for the `fragment`
        impossible_velocity_threshold : float
            If the velocity needed to link two fragments is higher than this threshold
            the identiy of one of the fragments is considerd to be wrong as it would be
            physically impossible for an animal to move so much. See `video.velocity_threshold`
            for each definition

        Returns
        -------
        candidate_identities_by_speed : nd.array
            Array with the identities that fullfill the `impossible_velocity_threshold`
            ordered from minimum to maximum velocity

        speed_of_candidate_identities : nd.array
            Array with the maximum velocity needed to link the given `fragment`
            with its neighbours assuming a given identity. Ordered from minimum to maximum
            velocity

        See Also
        --------
        Fragment
        get_candidate_identities_by_minimum_speed

        """
        P2_vector = fragment.P2_vector
        if len(non_available_identities) > 0:
            P2_vector[non_available_identities - 1] = 0
        if np.all(P2_vector == 0):
            candidate_identities_speed, _ = get_candidate_identities_by_minimum_speed(fragment, fragments, available_identities, impossible_velocity_threshold)
            return candidate_identities_speed
        else:
            if fragment.number_of_images == 1:
                random_threshold  = 1/fragment.number_of_animals
            else:
                random_threshold = 1/fragment.number_of_images
            return np.where(P2_vector > random_threshold)[0] + 1

    non_available_identities, \
    available_identities = get_available_and_non_available_identities(fragment)
    if len(available_identities) == 1:
        candidate_id = list(available_identities)[0]
    else:
        candidate_identities_speed, \
        speed_of_candidate_identities = get_candidate_identities_by_minimum_speed(fragment,
                                                                    fragments,
                                                                    available_identities,
                                                                    impossible_velocity_threshold)
        candidate_identities_P2 = get_candidate_identities_above_random_P2(fragment,
                                                                    fragments,
                                                                    non_available_identities,
                                                                    available_identities,
                                                                    impossible_velocity_threshold)
        candidate_identities = []
        candidate_speeds = []
        for candidate_id, candidate_speed in zip(candidate_identities_speed, speed_of_candidate_identities):
            if candidate_id in candidate_identities_P2:
                candidate_identities.append(candidate_id)
                candidate_speeds.append(candidate_speed)
        if len(candidate_identities) == 0:
            candidate_id = 0
        elif len(candidate_identities) == 1:
            if candidate_speeds[0] < impossible_velocity_threshold:
                candidate_id = candidate_identities[0]
            else:
                candidate_id = 0
        elif len(candidate_identities) > 1:
            if len(np.where(candidate_speeds == np.min(candidate_speeds))[0]) == 1:
                if candidate_speeds[0] < impossible_velocity_threshold:
                    candidate_id = candidate_identities[0]
                else:
                    candidate_id = 0
            else:
                candidate_id = 0

    fragment._identity_corrected_solving_jumps = candidate_id


def compute_velocities_consecutive_fragments(neighbour_fragment_past,
                                                fragment,
                                                neighbour_fragment_future):
    """Compute velocities in the extremes of a `fragment` with respecto to its
    `neighbour_fragment_past` and `neighbour_fragment_future`

    Parameters
    ----------
    neighbour_fragment_past : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        past
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of overlapping
        blobs that are considered to be the same animal
    neighbour_fragment_future : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        future

    Returns
    -------
    velocities : list
        List with the velocity needed to link the fragment with its fragment in
        the past and in the future

    See Also
    --------
    Fragment

    """
    velocities = [np.nan, np.nan]
    if neighbour_fragment_past is not None:
        velocities[0] = fragment.compute_border_velocity(neighbour_fragment_past)
    if neighbour_fragment_future is not None:
        velocities[1] = neighbour_fragment_future.compute_border_velocity(fragment)
    return velocities


def get_fragment_with_same_identity(video, list_of_fragments, fragment, direction):
    """Get the `neighbour_fragment` with the same identity in a given `direction`

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and methods
        related to them
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of overlapping
        blobs that are considered to be the same animal
    direction : string
        If `direction` = `to_the_past` gets the `neighbour_fragment` in the past
        `direction` = `to_the_future` gets the `neighbour_fragment` in the future

    Returns
    -------
    neighbour_fragment : <Fragment object>
        `Fragment` object with the same identity in a given `direction`
    number_of_frames_in_direction : int
        Number of frames to find the `neighbour_fragment` from a given extreme
        of the `fragment`

    See Also
    --------
    Fragment

    """
    neighbour_fragment = None
    number_of_frames_in_direction = 0
    frame_number = fragment.start_end[0] if direction == 'to_the_past' else fragment.start_end[1]

    while neighbour_fragment is None\
        and (frame_number > 0 and frame_number < video.number_of_frames):
        neighbour_fragment = fragment.get_neighbour_fragment(list_of_fragments.fragments,
                                        direction,
                                        number_of_frames_in_direction = number_of_frames_in_direction)
        number_of_frames_in_direction += 1
        frame_number += -1 if direction == 'to_the_past' else 1

    return neighbour_fragment, number_of_frames_in_direction


def compute_neighbour_fragments_and_velocities(video, list_of_fragments, fragment):
    """Computes the fragments with the same identities to the past and to the
    future of a given `fragment` and gives the velocities at the extremes of the
    current `fragment`

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and methods
        related to them
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of overlapping
        blobs that are considered to be the same animal

    Returns
    -------
    neighbour_fragment_past : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        past
    neighbour_fragment_future : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        future
    velocities_between_fragments : nd.array
        Velocities needed to connect the current fragment to its consecutive
        fragments in the past and in the future.

    See Also
    --------
    get_fragment_with_same_identity
    compute_velocities_consecutive_fragments

    """
    neighbour_fragment_past, \
    number_of_frames_in_past = get_fragment_with_same_identity(video,
                                                            list_of_fragments,
                                                            fragment,
                                                            'to_the_past')
    neighbour_fragment_future, \
    number_of_frames_in_future = get_fragment_with_same_identity(video,
                                                                list_of_fragments,
                                                                fragment,
                                                                'to_the_future')
    velocities = compute_velocities_consecutive_fragments(neighbour_fragment_past,
                                                            fragment,
                                                            neighbour_fragment_future)
    velocities_between_fragments = np.asarray(velocities) / np.asarray([number_of_frames_in_past, number_of_frames_in_future])

    return neighbour_fragment_past, neighbour_fragment_future, velocities_between_fragments


def correct_impossible_velocity_jumps_loop(video, list_of_fragments, scope = None):
    """Checks whether the velocity needed to join two consecutive fragments with
    the same identity is consistent with the typical velocity of the animals in
    the video (`video.velocity_threshold`). If the velocity is not consistent the
    identity of one of the fragments is reassigned. The check is performed from the
    `video.first_frame_first_global_fragment` to the past or to the future according
    to the `scope`

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and methods
        related to them
    scope : string
        If `scope` = `to_the_past` the check is performed to the past and if
        `scope` = `to_the_future` the check is performed to the future.

    See Also
    --------
    Fragment
    compute_neighbour_fragments_and_velocities
    compute_velocities_consecutive_fragments
    reassign

    """
    fragments_in_direction = list_of_fragments.get_ordered_list_of_fragments(scope, video.first_frame_first_global_fragment[video.accumulation_trial])
    impossible_velocity_threshold = video.velocity_threshold

    for fragment in tqdm(fragments_in_direction, desc = 'Correcting impossible velocity jumps ' + scope):
        if fragment.is_an_individual and fragment.assigned_identities[0] != 0:
            neighbour_fragment_past, \
            neighbour_fragment_future, \
            velocities_between_fragments = compute_neighbour_fragments_and_velocities(video, list_of_fragments, fragment)

            if all(velocity > impossible_velocity_threshold for velocity in velocities_between_fragments):
                if neighbour_fragment_past.identity_is_fixed or neighbour_fragment_future.identity_is_fixed:
                    reassign(fragment, list_of_fragments.fragments, impossible_velocity_threshold)
                else:
                    neighbour_fragment_past_past = neighbour_fragment_past.get_neighbour_fragment(list_of_fragments.fragments, 'to_the_past')
                    velocity_in_past = compute_velocities_consecutive_fragments(neighbour_fragment_past_past, neighbour_fragment_past, fragment)[0]
                    neighbour_fragment_future_future = neighbour_fragment_future.get_neighbour_fragment(list_of_fragments.fragments, 'to_the_future')
                    velocity_in_future = compute_velocities_consecutive_fragments(fragment, neighbour_fragment_future, neighbour_fragment_future_future)[1]
                    if velocity_in_past < impossible_velocity_threshold or velocity_in_future < impossible_velocity_threshold:
                        reassign(fragment, list_of_fragments.fragments, impossible_velocity_threshold)
            elif velocities_between_fragments[0] > impossible_velocity_threshold:
                if neighbour_fragment_past.identity_is_fixed:
                    reassign(fragment, list_of_fragments.fragments, impossible_velocity_threshold)
                else:
                    reassign(neighbour_fragment_past, list_of_fragments.fragments, impossible_velocity_threshold)
            elif velocities_between_fragments[1] > impossible_velocity_threshold:
                if neighbour_fragment_future.identity_is_fixed:
                    reassign(fragment, list_of_fragments.fragments, impossible_velocity_threshold)
                else:
                    reassign(neighbour_fragment_future, list_of_fragments.fragments, impossible_velocity_threshold)

def correct_impossible_velocity_jumps(video, list_of_fragments):
    """Corrects the parts of the video where the velocity of any individual is
    higher than a particular velocty threshold given by `video.velocity_threshold`.
    This check is done from the `video.first_frame_first_global_fragment` to the
    past and to the future

    Parameters
    ----------
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and methods
        related to them

    See Also
    --------
    correct_impossible_velocity_jumps_loop

    """
    correct_impossible_velocity_jumps_loop(video, list_of_fragments, scope = 'to_the_past')
    correct_impossible_velocity_jumps_loop(video, list_of_fragments, scope = 'to_the_future')



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H., de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P: gonzalo.polavieja@neuro.fchampalimaud.org)

import sys

import numpy as np
from confapp import conf


if sys.argv[0] == 'idtrackeraiApp.py' or 'idtrackeraiGUI' in sys.argv[0]:
    from kivy.logger import Logger
    logger = Logger
else:
    import logging
    logger = logging.getLogger("__main__.get_predictions")

class GetPrediction(object):
    """Manages the inference of the identities of a set of images

    Attributes
    ----------

    data_set : <DataSet object>
        Object containing the images whose labels has to be predicted
        (see :class:`~get_data.DataSet`)
    softmax_probs : ndarray
        Array of shape [number of images, number of animals] with the softmax vectors
        for every image in the `data_set`
    predictions : ndarray
        Array of shape [number of images, 1] with the predictions computed as the argmax
        of the softmax vector of every image (in dense format).
    _fc_vectors : ndarray
        Array of shape [number of images, number of neurons in second to last fully connected layer]
    batch_size : int
        Size of the batch to send the images through the network to get the predictions

    """
    def __init__(self, data_set):
        # Data set
        self.data_set = data_set
        self._softmax_probs = []
        self._predictions = []
        self._predictions_KNN = []
        self._fc_vectors = []
        self.batch_size = conf.BATCH_SIZE_PREDICTIONS_IDCNN

    @property
    def softmax_probs(self):
        return self._softmax_probs

    @property
    def predictions(self):
        return self._predictions

    def next_batch(self, batch_size):
        """Return the next `batch_size` examples from this data set.

        Parameters
        ----------
        batch_size : int
            Number of images to get for the next batch

        Returns
        -------
        images : ndarray
            Array of shape [batch_size, height, width, channels] with the images
            of the batch
        """
        start = self._index_in_epoch
        self._index_in_epoch += batch_size
        end = self._index_in_epoch
        return self.data_set.images[start:end]

    def get_predictions_softmax(self, batch_operation):
        """Runs a `batch_operation` (typically :meth:`~id_CNN.predict`)

        Parameters
        ----------
        batch_operation : func
            Function to be run in the epoch

        """
        self._index_in_epoch = 0
        while self._index_in_epoch < self.data_set._num_images:
            softmax_probs_batch, predictions_batch = batch_operation(self.next_batch(self.batch_size))
            self._softmax_probs.append(softmax_probs_batch)
            self._predictions.append(predictions_batch)
        self._softmax_probs = np.concatenate(self._softmax_probs, axis = 0)
        self._predictions = np.concatenate(self._predictions, axis = 0)
